from typing import Optional, Callable, Dict 
import chess, chess.pgn

def warnOfDanger(gameNode : chess.pgn.GameNode, log : Optional[Callable[[str], None]] = None,  depth : int = 0) -> Optional[Dict[chess.Square, int]]:
 ''' Warns, if certain pieces are in danger
 
Basis are the material scores from the attackers point of view, 
i.e. positive scores show an advantage of the attacker

:param gameNode: game node to be analysed
:returns: a dictionary of square/score pairs
  '''
 pieceScoreDict = {
  chess.PAWN : 1, 
  chess.KNIGHT : 3, 
  chess.BISHOP : 3, 
  chess.ROOK : 5, 
  chess.QUEEN : 9, 
  chess.KING : 1000
 }
 originalBoard = gameNode.board()
 fen = originalBoard.fen()
 if originalBoard.is_game_over():
  return None
 if log is not None and depth > 0:
  log(' move: {} (next_isWhite = {}, fen = {} ------------------------------'.format((gameNode.ply()+1)/2, originalBoard.turn, fen))
 square2ScoreDict : Dict[chess.Square, int] = dict()
 if originalBoard.is_check():
  square2ScoreDict[originalBoard.king(originalBoard.turn)] = - pieceScoreDict[chess.KING]
  if log is not None:
   log(' check detected!') 
  return square2ScoreDict
 originalBoard.push(chess.Move.null())
 for piece in chess.PIECE_TYPES:
  sList = list(originalBoard.pieces(piece, not originalBoard.turn))
  for sSquare in sList:
   board = originalBoard.copy()
   scoreList = list()
   sign = -1
   if log is not None and depth > 0:
    log('  square = {}, piece = {}'.format(chess.square_name(sSquare), chess.piece_name(piece))) 
   # 1. assign material scores to each attack and reply
   while True:
    minScore = 10000
    minSquare = None
    for actSquare in board.attackers(board.turn, sSquare):
     actPiece = board.piece_type_at(actSquare)
     # 2. promote to queen, if applicable
     if actPiece == chess.PAWN and \
         ((board.turn and chess.square_rank(sSquare) == 7) \
       or ((not board.turn) and chess.square_rank(sSquare) == 0)):
      promotion = chess.QUEEN
     else:
      promotion = None
     if chess.Move(actSquare, sSquare, promotion) in board.legal_moves:
      if minScore > pieceScoreDict[actPiece]:
       minScore = pieceScoreDict[actPiece]
       minSquare = actSquare
      if log is not None and depth > 0:
       log(' isAttacker = {}, square = {}, piece = {}'.format(sign == 1, chess.square_name(actSquare), chess.piece_name(board.piece_type_at(actSquare))))
     elif log is not None and depth > 0:
       log(' isAttacker = {}, square = {}, piece = {}, illegal move detected'.format(originalBoard.turn != board.turn, chess.square_name(actSquare), chess.piece_name(board.piece_type_at(actSquare))))
    if minScore == 10000:
     break
    scoreList.append(sign*minScore)
    board.push(chess.Move(minSquare, sSquare, promotion))
    sign = -sign
   # 2. no attackers
   if len(scoreList) >= 1:
    square2ScoreDict[sSquare] = pieceScoreDict[originalBoard.piece_type_at(sSquare)]
    if len(scoreList) > 1: 
     square2ScoreDict[sSquare] += sum(scoreList[:-1])
    if log is not None:
     log(' square = {}, totScore = {}'.format(chess.square_name(sSquare), square2ScoreDict[sSquare]))
 return square2ScoreDict

if __name__ == "__main__":
 import io
 from pgnParse import read_game

 def createGame(file):
  with open(file, mode = 'r',  encoding = 'utf-8') as f:
   newData = f.read()

  pgn = io.StringIO(newData)
  game = read_game(pgn)
  return game

 ps = "C:/Users/Reinh/OneDrive/Dokumente/Schach/ps210105.pgn"
 game = createGame(ps)
 gameNode = game.next()
 ply = 2
 while gameNode is not None:
  print(' move {}: {} ------------------------------'.format(ply/2, gameNode.move.uci()))
  square2ScoreDict = warnOfDanger(gameNode, log = print, depth = 0)
  gameNode = gameNode.next()
  if len(square2ScoreDict) > 0:
   print(' square2ScoreDict: ------------------------------')
   for square, score in square2ScoreDict.items():
    print('  {}: {}'.format(chess.square_name(square), score))
  ply += 1
 print('completed ------------------------------')



from typing import Optional,  Callable,  List
import sys, os, os.path

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import MzChess

if MzChess.useQt5:
 from PyQt5 import QtWidgets, QtGui, QtCore
 from PyQt5 import uic
else:
 from PyQt6 import QtWidgets, QtGui, QtCore
 from PyQt6 import uic

import chess, chess.pgn
import MzChess
import AboutDialog
from installLeipFont import installLeipFont

class BuildFenClass(QtWidgets.QMainWindow):
 '''The *chessboard* is based on Qt's QGraphicsView.
 '''

 def __init__(self, parent = None) -> None:
  super(BuildFenClass, self).__init__(parent)
  installLeipFont()
  fileDirectory = os.path.dirname(os.path.abspath(__file__))
  uic.loadUi(os.path.join(fileDirectory, 'buildFen.ui'), self)

  self.pgm = 'FEN-Builder'
  self.version = MzChess.__version__
  self.dateString = MzChess.__date__

  self.helpIndex = QtCore.QUrl('https://en.wikipedia.org/wiki/Forsyth%E2%80%93Edwards_Notation')
 
  sbText = self.statusBar().font()
  sbText.setPointSize(12)

  self.board = chess.Board()

  self.msgBox = QtWidgets.QMessageBox()
  self.msgBox.setIcon(QtWidgets.QMessageBox.Icon.Critical)
  self.msgBox.setWindowTitle("Error ...")

  self.infoLabel = QtWidgets.QLabel()
  self.infoLabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignLeft)
  self.infoLabel.setFont(sbText)
  self.statusBar().addPermanentWidget(self.infoLabel, 150)
 
  sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Preferred)
  sizePolicy.setHorizontalStretch(0)
  sizePolicy.setVerticalStretch(0)
  sizePolicy.setHeightForWidth(self.castlingGroupBox.sizePolicy().hasHeightForWidth())
  self.castlingGroupBox.setSizePolicy(sizePolicy)
  self.wkCheckBox = self. _addCastlingBox(chess.WHITE, True)
  self.wqCheckBox = self. _addCastlingBox(chess.WHITE, False)
  self.bkCheckBox = self. _addCastlingBox(chess.BLACK, True)
  self.bqCheckBox = self. _addCastlingBox(chess.BLACK, False)
  
  self.aboutDialog = AboutDialog.AboutDialog()
  self.aboutDialog.setup(
   pgm = self.pgm, 
   version = self.version, 
   dateString = self.dateString)

 def setup(self) -> None:
  self.selectionBox.setup()
  self.placementBoard.setup(self)
  self._resetFen()
  
 def notifyError(self, str : str) -> None:
  self.msgBox.setText(str)
  self.msgBox.exec()

 @QtCore.pyqtSlot(str)
 def notify(self, str : str) -> None:
  self.infoLabel.setText(str)
  self.infoLabel.update()
  QtWidgets.QApplication.processEvents()
  
 def _addCastlingBox(self, color : chess.Color, kingside : bool ) -> None:
  box = QtWidgets.QCheckBox(self.castlingGroupBox)
  box.toggled.connect(self.on_castlingCheckBox_toggled)
  if color == chess.BLACK:
   box.setStyleSheet('background-color: rgb(0, 0, 0);\ncolor: rgb(255, 255, 255);')
  if kingside:
   box.setText('O-O')
  else:
   box.setText('O-O-O')
  box.setChecked(False)
  self.castlingGroupBox.layout().addWidget(box)
  return box

 def _epList(self) -> None: 
  if self.board.turn:
   rank = 4
   sgn = 1
  else:
   rank = 3
   sgn = -1
  square2PieceList = sorted(self.board.piece_map().items())
  lastSquare = None
  lastColor = None
  sqList = list()
  for square, piece in square2PieceList:
   if piece.piece_type == chess.PAWN and chess.square_rank(square) == rank:
    if lastSquare != square - 1 or piece.color == lastColor:
     lastSquare = None
    if lastSquare is None:
     lastSquare = square
     lastColor = piece.color
     continue
    if piece.color == self.board.turn:
     target = chess.square_name(lastSquare+sgn*8)
    else:
     target = chess.square_name(square+sgn*8)
    if target not in sqList:
     sqList.append(target)
  return sqList
 
 def _resetFen(self) -> None:
  self.placementBoard.resetPosition()
  if self.board.turn:
   self.wRadioButton.click()
  else:
   self.bRadioButton.click()
  validWK = self.board.kings & chess.BB_E1 > 0
  self.wkCheckBox.setChecked(validWK and self.board.rooks & chess.BB_H1 > 0)
  self.wqCheckBox.setChecked(validWK and self.board.rooks & chess.BB_A1 > 0)
  validBK = self.board.kings & chess.BB_E8 > 0
  self.bkCheckBox.setChecked(validBK and self.board.rooks & chess.BB_H8 > 0)
  self.bqCheckBox.setChecked(validBK and self.board.rooks & chess.BB_A8 > 0)
  self.enPassantListWidget.clear()
  self.enPassantListWidget.addItem('-')
  for target in self._epList():
   self.enPassantListWidget.addItem(target)
  self.moveSpinBox.setValue(self.board.halfmove_clock)
  self.clockSpinBox.setValue(self.board.fullmove_number)
  self.notify(self.board.fen())

 @QtCore.pyqtSlot()
 def on_actionCopy_triggered(self):
  try:
   MzChess.checkFEN(self.board)
   fen = self.board.fen()
   QtWidgets.QApplication.clipboard().setText(fen)
  except ValueError as err:
   self.notifyError('Improper FEN {}:\n{}'.format(fen, str(err)))
   
 @QtCore.pyqtSlot()
 def on_actionPaste_triggered(self):
  fen = QtWidgets.QApplication.clipboard().text()
  try:
   MzChess.checkFEN(fen, allowIncompleteBoard = True)
   self.board.set_fen(fen)
   self._resetFen()
  except ValueError as err:
   self.notifyError('Improper FEN {}:\n{}'.format(fen, str(err)))

 @QtCore.pyqtSlot()
 def on_fenChanged(self):
  self.notify(self.board.fen())
   
 @QtCore.pyqtSlot()
 def on_actionReset_Board_triggered(self):
  self.board.reset()
  self._resetFen()
  
 @QtCore.pyqtSlot()
 def on_actionClear_Board_triggered(self):
  self.board.clear()
  self._resetFen()
  
 @QtCore.pyqtSlot()
 def on_wRadioButton_clicked(self):
  if self.board.turn != chess.WHITE:
   self.board.turn = chess.WHITE
   self._resetFen()

 @QtCore.pyqtSlot()
 def on_bRadioButton_clicked(self):
  if self.board.turn != chess.BLACK:
   self.board.turn = chess.BLACK
   self._resetFen()
  
 @QtCore.pyqtSlot(bool)
 def on_castlingCheckBox_toggled(self, checked):
  if    'wkCheckBox' not in vars(self) \
    or 'wqCheckBox' not in vars(self) \
    or 'bkCheckBox' not in vars(self) \
    or 'bqCheckBox' not in vars(self):
    return
  castlingString = ''
  if self.wkCheckBox.isChecked():
   castlingString += 'K'
  if self.wqCheckBox.isChecked():
   castlingString += 'Q'
  if self.bkCheckBox.isChecked():
   castlingString += 'k'
  if self.bqCheckBox.isChecked():
   castlingString += 'q'
  if len(castlingString) == 0:
   castlingString = '-'
  self.board.set_castling_fen(castlingString)
  self.notify(self.board.fen())
  
 @QtCore.pyqtSlot(QtWidgets.QListWidgetItem)
 def on_enPassantListWidget_itemClicked(self, enPassantItem):
  enPassantSqSymbol = enPassantItem.text()
  if enPassantSqSymbol != '-':
   self.board.ep_square = chess.parse_square(enPassantItem.text())
  else:
   self.board.ep_square = None
  self.notify(self.board.fen())

 @QtCore.pyqtSlot(int)
 def on_moveSpinBox_valueChanged(self, moveNumber):
  self.board.fullmove_number = moveNumber
  self.notify(self.board.fen())
  
 @QtCore.pyqtSlot(int)
 def on_clockSpinBox_valueChanged(self, halfmoveClock):
  self.board.halfmove_clock = halfmoveClock
  self.notify(self.board.fen())

 @QtCore.pyqtSlot()
 def on_actionAbout_triggered(self):
  self.notify('')
  self.aboutDialog.exec()
 
 @QtCore.pyqtSlot()
 def on_actionHelp_triggered(self):
  self.notify('')
  QtGui.QDesktopServices.openUrl(self.helpIndex)

class ChessGroupBox(QtWidgets.QGroupBox):
 leipzigEncodeDict = {
   'P' : 'p', 'N' : 'n', 'B' : 'b', 'R' : 'r', 'Q' : 'q', 'K' : 'k',
   'p' : 'o', 'n' : 'm', 'b' : 'v', 'r' : 't', 'q' : 'w', 'k' : 'l',
 }
 colorName = ['black', 'white']

 def __init__(self, parent = None) -> None:
  super(ChessGroupBox, self).__init__(parent)
  self.font = QtGui.QFont()
  self.font.setFamily("Chess Leipzig")
  self.font.setPointSize(24)
  self.gridLayout = QtWidgets.QGridLayout(self)
  self.gridLayout.setContentsMargins(5, 5, 5, 5)
  self.squareSize = 40
  
class SelectionBox(ChessGroupBox):

 def __init__(self, parent = None) -> None:
  super(SelectionBox, self).__init__(parent)
  self.button2PieceDict = dict()
  self.selectedPiece = None
  self.selectedButton = None

 def setup(self):
  self.pushButtonList = list()
  for s in 'KkQqRrBbNnPp':
   self._addButton(chess.Piece.from_symbol(s))
  self._addButton(None)
  
 def getButton(self, piece = Optional[chess.Piece]):
  for button, actPiece in self.button2PieceDict.items():
   if actPiece == piece:
    return button
   
 def _addButton(self, piece : Optional[chess.Piece] ) -> None:
  pushButton = QtWidgets.QPushButton(self)
  pushButton.setMinimumSize(QtCore.QSize(self.squareSize, self.squareSize))
  pushButton.setMaximumSize(QtCore.QSize(self.squareSize, self.squareSize))
  pushButton.setCheckable(True)
  pushButton.setFont(self.font)
  nPieces = len(self.button2PieceDict)
  self.button2PieceDict[pushButton] = piece
  if piece is None:
   pushButton.setChecked(True)
   pushButton.setText('x')
   pushButton.setShortcut('x')   
   pushButton.setToolTip('remove piece')
   self.selectedButton = pushButton
   self.gridLayout.addWidget(pushButton, nPieces // 2, nPieces % 2, 1, -1, QtCore.Qt.AlignmentFlag.AlignCenter)
  else:
   pushButton.setChecked(False)
   pushButton.setText(self.leipzigEncodeDict[piece.symbol()])
   if piece.color:
    sc = 'Shift+'
   else:
    sc = ''
   sc += piece.symbol().upper()
   pushButton.setShortcut(sc)   
   pushButton.setToolTip('{} {}'.format(self.colorName[piece.color], chess.piece_name(piece.piece_type)))
   self.gridLayout.addWidget(pushButton, nPieces // 2, nPieces % 2)
  pushButton.clicked.connect(self.on_clicked)
  pushButton.clicked.connect(self.on_clicked)
  self.pushButtonList.append(pushButton)
  
 @QtCore.pyqtSlot()
 def on_clicked(self):
  self.selectedButton.setChecked(False)
  self.selectedButton = self.sender()
  self.selectedPiece = self.button2PieceDict[self.selectedButton]
  self.sender().setChecked(True)

class PlacementBoard(ChessGroupBox):

 def __init__(self, parent = None) -> None:
  super(PlacementBoard, self).__init__(parent)
  self.gridLayout.setSpacing(0)
  self.button2SquareList = 64 * [None]
  self.flipped = False
 
 def setup(self, buildFenClass : BuildFenClass):
  self.buildFenClass = buildFenClass
  self.pushButtonList = list()
  for square in range(64):
   self._addButton(square)
  
 def resetPosition(self):
  pieceDict = self.buildFenClass.board.piece_map()
  for square, pushButton in enumerate(self.button2SquareList):
   if square in pieceDict:
    pushButton.setText(self.leipzigEncodeDict[pieceDict[square].symbol()])
   else:
    pushButton.setText('') 

 def setFlipped(self, flipped : bool):
  if self.flipped != flipped:
   for square in range(32):
    flippedSquare = square ^ 0x38
    tmp = self.button2SquareList[flippedSquare]
    self.button2SquareList[flippedSquare] = self.button2SquareList[square]
    self.button2SquareList[square] = tmp
    self._setBrushAndToolTip(self.button2SquareList[square], square)
    self._setBrushAndToolTip(self.button2SquareList[flippedSquare], flippedSquare)
    
 def _setBrushAndToolTip(self, pushButton : QtWidgets.QPushButton, square : chess.square) -> None:
  if (chess.square_rank(square) + self.flipped) % 2 == chess.square_file(square) % 2:
   pushButton.setStyleSheet("background-color: lightgray; \n;border: none;")  
  else:
   pushButton.setStyleSheet("background-color: white; \n;border: none;")  
  pushButton.setToolTip(chess.square_name(square))
  
 def _addButton(self, square : chess.square ) -> None:
  pushButton = QtWidgets.QPushButton(self)
  pushButton.setMinimumSize(QtCore.QSize(self.squareSize, self.squareSize))
  pushButton.setMaximumSize(QtCore.QSize(self.squareSize, self.squareSize))
  pushButton.setFont(self.font)
  pushButton.setText('')
  self._setBrushAndToolTip(pushButton, square)
  pushButton.clicked.connect(self.on_clicked)
  self.gridLayout.addWidget(pushButton, 7 - chess.square_rank(square), chess.square_file(square))
  self.button2SquareList[square] = pushButton
  self.pushButtonList.append(pushButton)

 @QtCore.pyqtSlot()
 def on_clicked(self):
  sendingButton = self.sender()
  square = self.button2SquareList.index(sendingButton)
  pieceDict = self.buildFenClass.board.piece_map()
  piece = self.buildFenClass.selectionBox.selectedPiece
  oldPiece = pieceDict.get(square, None)
  txt = ''
  if piece != oldPiece:
   if piece is None:
    pieceDict.pop(square, None)
   else:
    pieceDict[square] = piece
    txt = self.leipzigEncodeDict[piece.symbol()]
  try:
   newBoard = chess.Board(self.buildFenClass.board.fen())
   newBoard.set_piece_map(pieceDict)
   MzChess.checkFEN(newBoard, allowIncompleteBoard = True)
   sendingButton.setText(txt)
   self.buildFenClass.board.set_piece_map(pieceDict)
   self.buildFenClass.notify(self.buildFenClass.board.fen())
  except ValueError as err:
   self.buildFenClass.notifyError('Improper placemenent @ {}:\n{}'.format(chess.square_name(square), str(err)))
   
def showStatus(board):
 print('fen = {}'.format(board.fen()))
 for row in range(8):
  for col in range(8):
   chessSquare = chess.square(col, row)
   piece = board.piece_at(chessSquare)
   if piece is not None:
    print(' square = {}, name = {}, symbol = {}'.format(
     chessSquare, chess.square_name(chessSquare), piece.symbol()))
 for n, move in enumerate(board.move_stack):
  print('{}. {}'.format(n, move.uci()))

class MzClassApplication(QtWidgets.QApplication):
 def __init__(self, argv : List[str], notifyFct : Callable[[str], None] = print) -> None: 
  super(MzClassApplication, self).__init__(argv)
  self.notifyFct = notifyFct

 def notify(self, rec, ev):
  rc = super(MzClassApplication, self).notify(rec, ev)
  #self.notifyFct('{} -> Type(Event)= {}, handled = {}'.format(rec, ev.type(),  rc))
  return rc

def runFenBuilder(notifyFct : Optional[Callable[[str], None]] = None):
 os.chdir(os.path.expanduser('~'))
 if notifyFct is not None:
  qApp = MzClassApplication(sys.argv)
 else:
  qApp = QtWidgets.QApplication(sys.argv)
 chessMainWindow = BuildFenClass()
 chessMainWindow.show()
 chessMainWindow.setup()
 qApp.exec()

def _runFenBuilder():
 print('Hello, world')

if __name__ == "__main__":
 runFenBuilder(print)



from typing import Callable, TextIO, List

import pytest
import os, os.path

import MzChess
import chess

def test_FEN(pytestconfig):
 specifiedFEN = pytestconfig.getoption('FEN')
 if specifiedFEN is None:
  pytest.skip('FEN not specified')
  return
 MzChess.checkFEN(specifiedFEN)

def test_pgnLexer(pytestconfig):
 if 'l' not in pytestconfig.getoption('target'):
  pytest.skip('Lexer test not activated')
  return
 pgn = pytest.helpers.findPGNTextIO(pytestconfig)
 lexer = MzChess.PGNLexer(bufsize = 4096, debug =  pytestconfig.getoption('debug'), optimize = False)
 tok = lexer.newGame(pgn)
 lexer.dumps(tok = tok, notify = print)

def test_readGame(pytestconfig):
 def runNode(readFct : Callable, f : TextIO,  gameID : int):
  try:
   node = readFct(f)
   print('node #{}, fct = {} -> {}'.format(gameID, readFct, repr(node)))
  except Exception as error:
   node = None
   print('node #{}: fct = {} failed -> {}'.format(gameID,readFct, error))
  return node
  
 def cmpGameNodes(chessNode, newNode):
  if isinstance(chessNode, bool) or isinstance(newNode, bool):
   context = 'Skipped game: chess.pgn: {}, MzChess = {}'.format(chessNode, newNode)
   assert chessNode == newNode, context
  elif isinstance(chessNode, chess.pgn.Headers) or isinstance(newNode, chess.pgn.Headers):
   context = 'Headers: chess.pgn: {}, MzChess = {}'.format(chessNode, newNode)
   assert chessNode == newNode, context
  elif isinstance(chessNode, chess.Board) or isinstance(newNode, chess.Board):
   context = 'Board: chess.pgn: {}, MzChess = {}'.format(chessNode.fen(), newNode.fen())
   assert chessNode.fen() == newNode.fen(), context
  else:
   if isinstance(chessNode, chess.pgn.Game) and isinstance(newNode, chess.pgn.Game):
    context = 'Game Headers: chess.pgn: {}, MzChess = {}'.format(chessNode.headers, newNode.headers)
    assert chessNode.headers == newNode.headers, context
   else:
    context = 'Move: chess.pgn: {}, MzChess = {}'.format(repr(chessNode), repr(newNode))
    assert chessNode.move.uci() == newNode.move.uci(), context
   assert chessNode.comment == newNode.comment, 'Comment: chess.pgn: "{}", MzChess = "{}" => {}'.format(chessNode.comment, newNode.comment, context)
   assert chessNode.nags == newNode.nags, 'Nags: chess.pgn: {}, MzChess = {} => {}'.format(sorted(chessNode.nags), sorted(newNode.nags), context)
   assert len(chessNode.variations) == len(newNode.variations), 'nVariations: chess.pgn: {}, MzChess = {} => {}'.format(len(chessNode.variations), len(newNode.variations), context)
   for subNode1, subNode2 in zip(chessNode.variations, newNode.variations):
    cmpGameNodes(subNode1, subNode2)
    
 # ===============================================================================

 targets = pytestconfig.getoption('target')
 anyExecuted = False
 for target in targets:
  pgn = pytest.helpers.findPGNTextIO(pytestconfig)
  newRead2 = None
  oldRead2 = None
  if target == 'g':
   newRead1 = MzChess.read_game
   oldRead1 = chess.pgn.read_game
  elif target == 'h':
   newRead1 = MzChess.read_headers
   oldRead1 = chess.pgn.read_headers
  elif target == 'b':
   newRead1 = MzChess.read_board
   oldRead1 = MzChess.read_board2
  elif target == 's':
   newRead1 = MzChess.read_game
   oldRead1 = chess.pgn.read_game
   newRead2 = MzChess.skip_game
   oldRead2 = chess.pgn.skip_game
  else:
   continue
  
  anyExecuted = True
  gameList = list()
  print('Parse games using new read_game ----------------------------')
  gameID = 0
  log = list()
  node = False
  while node is not None:
   gameID += 1
   node = runNode(newRead1, pgn, gameID)
   gameList.append(node)
   if newRead2 is not None:
    runNode(newRead2, pgn, gameID)
  gameID -= 1
  if newRead2 is not None:
   gameID *= 2
  log.append('MzChess.read_game: {} games read'.format(gameID))
  del gameList[-1]
  if pytestconfig.getoption('compare') and len(gameList) > 0:
   pgn.seek(0)
   pgnGameList = list()
   print('Parse games using chess.pgn.read_game ----------------------------')
   gameID = 0
   nErrors = 0
   node = False
   while node is not None:
    gameID += 1
    node = runNode(oldRead1, pgn, gameID)
    pgnGameList.append(node)
    if oldRead2 is not None:
     runNode(oldRead2, pgn, gameID)
   gameID -= 1
   if oldRead2 is not None:
    gameID *= 2
   log.append('chess.read_game: {} games read'.format(gameID - 1))
   del pgnGameList[-1]
   if len(gameList) != len(pgnGameList):
    print('len(gameList) = {} != len(pgnGameList) = {}'.format(len(gameList), len(pgnGameList)))
   print('Comparing games ----------------------------')
   gameID = 0
   for node, chessNode in zip(gameList, pgnGameList):
    gameID += 1
    txt = 'game #{}: {}'.format(gameID, repr(node))
    try:
     cmpGameNodes(chessNode, node)
     print(' {} succeeded'.format(txt))
    except AssertionError as msg:
     nErrors += 1
     print(' {} failed \n -> {}'.format(txt, msg))
   print('Comparison failed for {} of {} games --------'.format(nErrors, gameID))
  print('\n'.join(log)) 
  
  if not anyExecuted:
   pytest.skip('read_game tests not activated')




import pytest
from pytestqt.qt_compat import qt_api

from PyQt6 import QtCore, QtWidgets

import chess
import MzChess

@pytest.fixture
def app(qtbot):
 fenBuilder = MzChess.BuildFenClass()
 fenBuilder.show()
 fenBuilder.setup()
 qtbot.addWidget(fenBuilder)
 return fenBuilder
 
def assertCheckedButton(selectionBox : MzChess.SelectionBox, button : QtWidgets.QPushButton) -> None:
 isChecked = button.isChecked()
 for actButton in selectionBox.button2PieceDict:
  if actButton is button:
   assert isChecked
  else:
   assert not actButton.isChecked()
   
def handleNotifyError(app, qtbot, expectedMessage : str):
 print('Notify Error expected, press OK to continue ...')
 assert isinstance(app.msgBox, QtWidgets.QMessageBox), 'QMessageBox expected, {} got'.format(app.msgBox)
 assert app.msgBox.text() == expectedMessage
 okButton = app.msgBox.button(QtWidgets.QMessageBox.StandardButton.Ok)
 assert isinstance(okButton, QtWidgets.QPushButton)
 if app.msgBox.isVisible():
  qtbot.mouseClick(okButton, QtCore.Qt.MouseButton.LeftButton, delay=1)

def test_selectionBox(app, qtbot):
 assert app.selectionBox.selectedPiece is None
 print('test_selectionBox: check mouse clicks')
 for button, piece in app.selectionBox.button2PieceDict.items():
  qtbot.mouseClick(button, QtCore.Qt.MouseButton.LeftButton)
  assert app.selectionBox.selectedButton is button
  assert app.selectionBox.selectedPiece is piece
  assertCheckedButton(app.selectionBox, button)
 return
 print('test_selectionBox: check shortcuts')
 for s in 'KQRBNP':
  qtbot.keyClicks(app.selectionBox, s)
  assert app.selectionBox.selectedPiece == chess.Piece.from_symbol(s.lower())
  qtbot.keyClicks(app.selectionBox, 'Shift+' + s)
  assert app.selectionBox.selectedPiece == chess.Piece.from_symbol(s)

def test_placementBoard(app, qtbot):
 assert app.selectionBox.selectedPiece is None
 app.wkCheckBox.setChecked(False)
 app.wqCheckBox.setChecked(False)
 app.bkCheckBox.setChecked(False)
 app.bqCheckBox.setChecked(False)
 for button in app.placementBoard.pushButtonList:
  qtbot.mouseClick(button, QtCore.Qt.MouseButton.LeftButton)
 button =  app.selectionBox.getButton(chess.Piece.from_symbol('P'))
 qtbot.mouseClick(button, QtCore.Qt.MouseButton.LeftButton)
 qtbot.mouseClick(app.placementBoard.pushButtonList[0], QtCore.Qt.MouseButton.LeftButton)
 handleNotifyError(app, qtbot, 'Improper placemenent @ a1:\nPawn on rank (row) 1 or 8.')

def test_exitFenBuilder():
 pytest.helpers.exitApp()



__description__ = \
"""
A PyQt6 based chess GUI

Versions:
 1.0.0      first version
 1.0.3      installation running
 1.0.4      bug fixes + end-of-game handling completed
 1.0.6      documentation on github.io + ico added
 1.0.8      setup corrected
 1.1.0      FEN-builder added and bug fixes
 1.2.0      tests established and bug fixes
 1.2.1      variant bug fixes
 1.2.2      bug fixes
 1.2.3      bug fixes
 1.3.0      migrated to uic.loadUI
 1.4.0      migrated to PyQt6, Open/Save issues removed
 1.5.0      migrated to combined PyQt6/PyQt5 operation
"""
__author__ = "Reinhard Maerz"
__date__ = "2022-07-28"
__version__ = "1.5.0"

__all__ = [
 'AboutDialog', 
 'AnnotateEngine', 'Annotator', 
 'ChessMainWindow', 'runMzChess', 
 'BuildFenClass', 'runFenBuilder', 
 'installLeipFont'
 'ChessEngine', 
 'ConfigureEngine', 'loadEngineSettings', 'saveEngineSettings', 
 'ConfigureEngineOptions',  
 'ECODatabase', 'TSVType', 
 'GameHeaderView', 'KeyType', 
 'GameListTableModel', 'GameListTableView', 
 'HelpBrowser', 
 'checkFEN','read_game', 'read_board', 'read_headers', 'skip_game', 'PGNLexer', 
 'QBoardViewClass', 'Piece', 'Game', 
 'ScorePlot', 
 'ButtonLine', 'ItemSelector', 'treeWidgetItemPos', 
 'QUCIEdit', 'UCIHighlighter', 
 'warnOfDanger'
]


import os.path, sys
try:
 import PyQt6
 useQt5 = False
except:
 import PyQt5
 useQt5 = True

sys.path.append(os.path.dirname(os.path.abspath(__file__)))

from .AboutDialog import AboutDialog
from .annotateEngine import AnnotateEngine,  Annotator
from .installLeipFont import installLeipFont
from .chessengine import ChessEngine
from .configureEngine import ConfigureEngine, loadEngineSettings, saveEngineSettings
from .configureEngineOptions import ConfigureEngineOptions
from .eco import ECODatabase, TSVType
from .gameheaderview import GameHeaderView, KeyType
from .gamelisttableview import GameListTableModel, GameListTableView
from .helpDialog import HelpBrowser
from .pgnParse import checkFEN, read_game, read_board, read_headers, skip_game, PGNLexer
from .qboardviewclass import QBoardViewClass, Piece, Game
from .scoreplotgraphicsview import ScorePlot
from .specialDialogs import ButtonLine, ItemSelector, treeWidgetItemPos
from .uciedit import QUCIEdit, UCIHighlighter
from .warnOfDanger import warnOfDanger
from .qbuildfen import BuildFenClass, SelectionBox, PlacementBoard, runFenBuilder
from .chessMainWindow import ChessMainWindow, runMzChess



#!/usr/bin/python3

import chess
import itertools
import sys


def main(arg, db):
    ret = 0
    prev_eco = ""
    prev_name = ""

    with open(arg) as f:
        for lno, line in zip(itertools.count(1), f):
            cols = line.rstrip("\n").split("\t")

            if len(cols) != 4:
                print(f"::error file={arg},line={lno}::expected 4 columns, got {len(cols)}")
                ret = 1
                continue

            if lno == 1:
                if cols != ["eco", "name", "fen", "moves"]:
                    print(f"::error file={arg},line={lno}::expected eco, name, fen, moves")
                    ret = 1
                continue

            eco, name, fen, moves = cols
            moves = moves.split(" ")

            if name.count(":") > 1:
                print(f"::warning file={arg},line={lno}::multiple ':' in name: {name}")

            if fen in db:
                print(f"::warning file={arg},line={lno}::duplicate fen: {db[fen]}")
            else:
                db[fen] = cols

            if eco < prev_eco:
                print(f"::warning file={arg},line={lno}::not ordered by eco ({eco} after {prev_eco})")
            elif (eco, name) < (prev_eco, prev_name):
                print(f"::warning file={arg},line={lno}::not ordered by name ({name!r} after {prev_name!r})")
            prev_eco = eco
            prev_name = name

            try:
                board = chess.Board()
                for move in moves:
                    board.push_uci(move)
                if fen != board.epd():
                    print(f"::error file={arg},line={lno}::mismatching fen, expected: {board.fen()}")
                    ret = 1
            except ValueError as err:
                print(f"::error file={arg},line={lno}::{err}")
                ret = 1

            for blacklisted in ["refused"]:
                if blacklisted in name.lower():
                    print(f"::warning file={arg},line={lno}::blacklisted word ({blacklisted!r} in {name!r})")

    return ret


if __name__ == "__main__":
    if len(sys.argv) <= 1:
        print(f"Usage: {sys.argv[0]} *.tsv")
        sys.exit(2)

    db = {}
    ret = 0
    for arg in sys.argv[1:]:
        ret = max(ret, main(arg, db))
    sys.exit(ret)



from typing import Dict, Any
import sys,  os,  os.path

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import MzChess

if MzChess.useQt5:
 from PyQt5 import QtWidgets, QtCore
 from PyQt5 import uic
else:
 from PyQt6 import QtWidgets, QtCore
 from PyQt6 import uic

class ConfigureEngineOptions(QtWidgets.QDialog):
 fileDialogOptions = QtWidgets.QFileDialog.Option.DontUseNativeDialog
 fileDialogFilters = QtCore.QDir.Filter.AllDirs \
                      | QtCore.QDir.Filter.Files \
                      | QtCore.QDir.Filter.NoDotAndDotDot \
                      | QtCore.QDir.Filter.Hidden
 readOnlyOptions = ['UCI_Chess960', 'UCI_AnalyseMode', 'MultiPV']
 checkState = {'false' : QtCore.Qt.CheckState.Unchecked, '0' : QtCore.Qt.CheckState.Unchecked, 
                      'true' :  QtCore.Qt.CheckState.Checked,  '1' : QtCore.Qt.CheckState.Unchecked}

 def __init__(self, parent = None):
  super(ConfigureEngineOptions, self).__init__(parent)
  fileDirectory = os.path.dirname(os.path.abspath(__file__))
  uic.loadUi(os.path.join(fileDirectory, 'configureEngineOptions.ui'), self)

  self.tableWidget.setColumnCount(3)  
  self.tableWidget.setHorizontalHeaderLabels(['','Option', 'Value'])
  self.tableWidget.horizontalHeader().setStretchLastSection(True)
  self.tableWidget.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectionBehavior.SelectRows)
  self.buttonBox.accepted.connect(self.accept)
  self.buttonBox.rejected.connect(self.reject)
 
 def run(self, engineOptionsDict : Dict[str, Dict[str, Any]], 
                   changedOptions : Dict[str, Any] = dict(), 
                   name : str = ''):
  if len(name) == 0:                  
   self.setWindowTitle('Configure Engines Options')
  else:
   self.setWindowTitle('Configure Options: {}'.format(name))
  self.tableWidget.setRowCount(len(engineOptionsDict))
  test = QtWidgets.QLabel()
  maxWidth = 0
  self.engineOptionsDict = engineOptionsDict
  self.optionsDict = dict()
  for key, value in changedOptions.items():
   self._toOptionsDict(key,  value)
  self.itemList = list()
  self.btnList = list()
  for row, key2Opts in enumerate(engineOptionsDict.items()):
   key, optDict = key2Opts
   maxWidth = max(maxWidth, test.fontMetrics().size(QtCore.Qt.TextFlag.TextSingleLine, key).width())
   nameLabel = QtWidgets.QLabel(str(key))
   self.btnList.append(None)
   type = optDict['type']
   if 'default' in optDict:
    if key in self.optionsDict:
     value = self.optionsDict[key]
    else:
     value = optDict['default'] 
   if type == 'check':
    item = QtWidgets.QCheckBox()
    item.setCheckState(self.checkState[value])
    item.stateChanged.connect(self.on_checkBox_stateChanged)
   elif type == 'spin':
    item = QtWidgets.QSpinBox()
    itemRange = optDict['range']
    item.setMinimum(itemRange.start)
    item.setMaximum(itemRange.stop - 1)
    item.setValue(int(value))
    item.valueChanged.connect(self.on_spinBox_changed)
   elif type == 'combo':
    item = QtWidgets.QComboBox()
    varList = optDict['varList']
    item.addItems(varList)
    item.setCurrentIndex(varList.index(value))
    item.currentTextChanged.connect(self.on_comboBox_changed)
   elif type == 'button':
    item = QtWidgets.QLabel('<no parameter>')
   elif type == 'string':
    if 'file' in key.lower() or 'path' in key.lower():
     browsePushButton = QtWidgets.QPushButton(self)
     browsePushButton.setIcon(self.style().standardIcon(QtWidgets.QStyle.StandardPixmap.SP_DialogSaveButton))
     self.tableWidget.setCellWidget(row, 0, browsePushButton)
     browsePushButton.clicked.connect(self.on_browsePushButton_clicked)
    self.btnList[-1] = browsePushButton
    item = QtWidgets.QLineEdit()
    item.setText(value)
    item.editingFinished.connect(self.on_string_editingFinished)
   else:
    raise ValueError('ConfigureEngineOptions/run: improper type = {}'.format(type))
   if self.btnList[-1] is None:
    self.tableWidget.setCellWidget(row, 0, QtWidgets.QLabel(''))
   self.tableWidget.setCellWidget(row, 1, nameLabel)
   self.tableWidget.setCellWidget(row, 2, item)
   self.itemList.append(item)
  self.tableWidget.horizontalHeader().resizeSection(1, maxWidth)
  self.tableWidget.horizontalHeader().resizeSection(0, self.tableWidget.rowHeight(0))
  self.optionsDict = dict()
  self.tableWidget.update()
  QtCore.QCoreApplication.processEvents()
  if self.exec():
   return self.optionsDict
  else:
   return changedOptions
   
 def _toOptionsDict(self, key,  value):
  if self.engineOptionsDict[key]['default'] != value:
   self.optionsDict[key] = value
  else:
   self.optionsDict.pop(key, None)
 
 @QtCore.pyqtSlot(int)
 def on_checkBox_stateChanged(self, state):
  item = self.sender()  
  row = self.itemList.index(item)
  key = self.tableWidget.cellWidget(row, 1).text()
  if state == QtCore.Qt.CheckState.Checked:
   value = 'true'
  else:
   value = 'false'
  self._toOptionsDict(key,  value)
   
 @QtCore.pyqtSlot(int)
 def on_spinBox_changed(self, value):
  item = self.sender()  
  row = self.itemList.index(item)
  key = self.tableWidget.cellWidget(row, 1).text()
  self._toOptionsDict(key,  value)

 @QtCore.pyqtSlot(str)
 def on_comboBox_changed(self, value):
  item = self.sender()  
  row = self.itemList.index(item)
  key = self.tableWidget.cellWidget(row, 1).text()
  self._toOptionsDict(key,  value)

 @QtCore.pyqtSlot()
 def on_string_editingFinished(self):
  item = self.sender()  
  row = self.itemList.index(item)
  key = self.tableWidget.cellWidget(row, 1).text()
  value = item.text()
  self._toOptionsDict(key,  value)
   
 @QtCore.pyqtSlot()
 def on_browsePushButton_clicked(self):
  item = self.sender()  
  row = self.btnList.index(item)
  keyLabel = self.tableWidget.cellWidget(row, 1)
  key = keyLabel.text()
  valueLabel = self.tableWidget.cellWidget(row, 2)
  value = valueLabel.text()
  fDialog = QtWidgets.QFileDialog()
  fDialog.setOptions(self.fileDialogOptions)
  fDialog.setFilter(self.fileDialogFilters)
  if 'file' in key.lower(): 
   fDialog.setFileMode(QtWidgets.QFileDialog.FileMode.ExistingFile)
   fDialog.setWindowTitle("Set File ...")
   if len(value) > 0:
    head, tail = os.path.split(value)
    fDialog.setDirectory(head)
    fDialog.selectFile(tail)
  else:
   fDialog.setFileMode(QtWidgets.QFileDialog.FileMode.Directory)
   fDialog.setWindowTitle("Set Directory ...")
   if len(value) > 0:
    head, tail = os.path.split(value)
    fDialog.setDirectory(head)
  if fDialog.exec():
   value = fDialog.selectedFiles()[0]
   valueLabel = self.tableWidget.cellWidget(row, 2)
   valueLabel.setText(value)
   self._toOptionsDict(key,  value)
 
 @QtCore.pyqtSlot()
 def on_cancelPushButton_clicked(self):
  self.done(False)
  
 @QtCore.pyqtSlot()
 def on_okPushButton_clicked(self):
  self.done(True)


if __name__ == "__main__":
 import sys
 os.chdir('C:/Users/Reinh/chess_engines')
 stockfish12 = 'stockfish_12_win_x64_bmi2/stockfish_20090216_x64_bmi2.exe'
 togaII4 = 'TogaII40/Windows/TogaII_40_intelPGO_x64.exe'
 komodo12 = 'komodo-12.1.1_5a8fc2/Windows/komodo-12.1.1-64bit-bmi2.exe'
 fritz17 = 'Fritz17/Fritz17.exe'
 lcZero26 = 'lc0-v0.26.3-windows-gpu-opencl/lc0.exe'
 raubfischGTZ23 = 'RaubfischX44_and_GTZ23/GTZ23/RaubfischGTZ23_nn_sl-bmi2.exe'

 executable = stockfish12
 engineOptionsDict = {
 'Debug Log File': {'type': 'string', 'default': ''}, 
 'Contempt': {'type': 'spin', 'default': 24, 'range': range(-100, 101)}, 
 'Analysis Contempt': {'varList': ['Off', 'White', 'Black', 'Both'], 'type': 'combo', 'default': 'Both'}, 
 'Threads': {'type': 'spin', 'default': 1, 'range': range(1, 513)}, 
 'Hash': {'type': 'spin', 'default': 16, 'range': range(1, 33554433)}, 
 'Clear Hash': {'type': 'button'}, 
 'Ponder': {'type': 'check', 'default': 'false'}, 
 'MultiPV': {'type': 'spin', 'default': 1, 'range': range(1, 501)}, 
 'Skill Level': {'type': 'spin', 'default': 20, 'range': range(0, 21)}, 
 'Move Overhead': {'type': 'spin', 'default': 10, 'range': range(0, 5001)}, 
 'Slow Mover': {'type': 'spin', 'default': 100, 'range': range(10, 1001)}, 
 'nodestime': {'type': 'spin', 'default': 0, 'range': range(0, 10001)}, 
 'UCI_Chess960': {'type': 'check', 'default': 'false'}, 
 'UCI_AnalyseMode': {'type': 'check', 'default': 'false'}, 
 'UCI_LimitStrength': {'type': 'check', 'default': 'false'}, 
 'UCI_Elo': {'type': 'spin', 'default': 1350, 'range': range(1350, 2851)}, 
 'UCI_ShowWDL': {'type': 'check', 'default': 'false'}, 
 'SyzygyPath': {'type': 'string', 'default': '<empty>'}, 
 'SyzygyProbeDepth': {'type': 'spin', 'default': 1, 'range': range(1, 101)}, 
 'Syzygy50MoveRule': {'type': 'check', 'default': 'true'}, 
 'SyzygyProbeLimit': {'type': 'spin', 'default': 7, 'range': range(0, 8)}, 
 'Use NNUE': {'type': 'check', 'default': 'true'}, 
 'EvalFile': {'type': 'string', 'default': 'nn-82215d0fd0df.nnue'}} 

 changedOptions = { 'UCI_Elo' : 1410, 'Ponder' :  'true'}

 app = QtWidgets.QApplication([])
 dialog = ConfigureEngineOptions()
 dialog.run(engineOptionsDict, changedOptions = changedOptions)
 
 sys.exit(app.exec())



import platform
import shutil
import sys
import os, os.path

# Try using setuptools first, if it's installed
from setuptools import setup
from packaging import version

package = 'MzChess'
fileDirectory = os.path.dirname(os.path.abspath(__file__))
packageDirectory = os.path.join(fileDirectory, package)
sys.path.insert(0, packageDirectory)

with open(os.path.join(packageDirectory,'readme.rst'), 'r', encoding = 'utf-8') as f:
 long_description = f.read()
 
import MzChess
pkgVersion = MzChess.__version__

leipfontFile = os.path.join(packageDirectory, 'pieces', 'LEIPFONT.ttf')

if platform.system() == 'Linux':
 # On Linux systems PyQt must be installed using a package manager
 install_requires = list()
 try:
  import PyQt6.QtCore
  foundPyQtVersion = PyQt6.QtCore.PYQT_VERSION_STR
  if version.parse(foundPyQtVersion) < version.parse('6.2'): 
   print('Upgrade to PyQt6>=6.2 using your Linux Package Manager')
  try:
   import PyQt6.QtSvgWidgets
  except:
   print('Install PyQt6.QtSvgWidgets using your Linux Package Manager')
   quit()
  try:
   import PyQt6.QtCharts
  except:
   print('Install PyQt6.QtCharts using your Linux Package Manager')
   quit()
# check PyQt5
 except:
  import PyQt5.QtCore
  foundPyQtVersion = PyQt5.QtCore.PYQT_VERSION_STR
  if version.parse(foundPyQtVersion) < version.parse('5.11'): 
   print('Upgrade to PyQt6>=6.2 or PyQt5>=5.11 using your Linux Package Manager')
  try:
   import PyQt5.QtSvg
  except:
   print('Install PyQt5.QtSvg using your Linux Package Manager')
   quit()
  try:
   import PyQt5.QtCharts
  except:
   print('Install PyQt5.QtCharts using your Linux Package Manager')
   quit()
else:
 pyQt6Version = '6.2.0'
 install_requires = ['PyQt6>={}'.format(pyQt6Version), 'PyQt6-Charts>={}'.format(pyQt6Version)]

install_requires += ['chess>=1.4', 'ply>=3.11']

# Required to ensure a clean environment
shutil.rmtree(os.path.join(fileDirectory, 'build'), ignore_errors = True)

# Need to add all dependencies to setup as we go!
setup(name = package,
  url = 'https://github.com/ReinhardM-dev/MzChess', 
  project_urls={ 'Documentation': 'https://reinhardm-dev.github.io/MzChess' }, 
  version = pkgVersion,
  packages = [package],
  options={'bdist_wheel':{'universal':True}},
  package_data = {package: 
    ['*.txt', '*.ui', '*.gpl3', '*.rst', '*.png',  
     'books/*.txt', 
     'eco/*.fsc', 'eco/*.tsv', 'eco/*.txt', 'eco/*.md', 'eco/bin/*',
     'pieces/*.svg', 'pieces/*.ttf', 
     'training/*/*.pgn'] }, 
  description = 'A Chess GUI using PyQt6',
  long_description = long_description, 
  long_description_content_type="text/x-rst",
  entry_points={ 'gui_scripts': 
                         ['buildFen = {}:runFenBuilder'.format(package),
                          'mzChess = {}:runMzChess'.format(package)] }, 
  author  ='Reinhard Maerz',
  python_requires = '>=3.7', 
  install_requires = install_requires,
  setup_requires=['wheel'], 
  classifiers = [
    'Programming Language :: Python', 
    'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3 :: Only',
    'Development Status :: 4 - Beta', 
    'Natural Language :: English', 
    'Topic :: Games/Entertainment :: Board Games'])





'''
Main Window of the Chess GUI
================================
|MainWindow| 

The main Window consists of 4 parts:

 * On the left, the *game board* with the *material* and *next turn* labels. The *next turn* may 4 colors ``white``, ``black``, ``red`` (mate), ``yellow`` (draw)
 * On the right, the TAB widget with the
    
    * *Game* tree widget displaying the body of the actual game
    * *Headers* list widget displaying the header of the actual game
    * *Database* list widget displaying all loaded games
    * *Score* plot displaying engine and material scores
    * *Log* edit displaying communication with the engine in use
    
 * On the bottom, the status bar with the labels

    * *info/error* displaying information or errors
    * *engine* displaying the engine in use
    * *hint/score* displaying the hint and score computed by the engine
    * *ECO code* displaying the eco code of the current opening, the hint shows the full opening string
    * *square* displaying the square under the cursor within the board

Main Window Menu
================================
On the top, the menu with

 * *File* menu handling files in the Portable Game Notation (`PGN`_) and pickled PGN (*PPNG*) formats

    * *Encoding* sub-menu to set encoding for opening/saveing PNG-format
    * *Open DB ...* action to open a PPGN- or PGN-file and replace the current database
    * *Recent* sub-menu with the recent PPGN- or PGN-files
    * *Append to DB ...* action to open a PPGN- or PGN-file and append to the current database
    * *Save DB ...* action to save the whole database as a PPGN- or PGN-file 
    * *Save Game ...* action to save the actual game as a PGN-file 
    * *Exit* action to terminate the GUI

 * *Edit* menu with obvious functionality with the exception of
 
    * *Copy/Paste PGN*, i.e. the actual game is copied to or pasted from the clipboard
    * *Copy/Paste FEN*, i.e. the actual position is copied to or pasted from the clipboard
    * *Promote/Demote Variant* promote/demote the variant selected in the Game TAB
    * *Promote Variant to Main* promote the variant selected in the Game TAB to mainline
    * *Delete Variant* delete the variant selected in the Game TAB

 * *Database* menu with obvious functionality with the exception of
 
    * *Remove Games* removes selected games in the *Database* TAB

 * *Game* menu with obvious functionality with the exception of
 
    * *Select Header Elements ...* opens a dialog to add/remove header elements according to the `PGN`_ standard
    * *Show Move Options* shows by left-button selecting a square the weighted options 
    * *Warn of Danger* shows squares attacked by the opponent 

 * *Engine* menu for handling Universal Chess Interface (`UCI`_) engines with
 
    * *Select Engine* sub-menu to select the engine for hints/scores and annotations
    * *Search Depth* sub-menu to set the search depth of the selected engine
    * *Annotate Last Move* annotates the move leading to actual position
    * *Annotate All* annotates the actual game
    * *# Annotations* defines the number of variants to be suggested in case of a blunder
    * *Annotate Variants* defines the number of half moves (PLY) to be shown in variants 
    * *Show Scores* toggle actions enables the *score* part of the *hint/score* label of the status bar
    * *Show Hints* toggle actions enables the *hint* part of the *hint/score* label of the status bar
    * *Configure ...* opens a dialog to add/remove/configure engines
    * *Debug* logs the communication with engine in the *Log* TAB

Keyboard and Mouse Contol
================================

To allow for a user-friendly usage, the game can be controlled by keyboard and/or mouse:

.. csv-table:: Game Control by Keyboard/Mouse/Mouse Wheel
   :header: "Key/Wheel", "Board", "Game-TAB", "Description"
   :widths: 30, 10, 10, 50
   
   :kbd:`up` | :kbd:`scroll-up`       , :kbd:`✔`, :kbd:`✔`, goto last move
   :kbd:`down` | :kbd:`scroll-down`, :kbd:`✔`, :kbd:`✔`, goto next move
   :kbd:`left`                               ,              , :kbd:`✔`, step into variant
   :kbd:`right`                             ,              , :kbd:`✔`, step out of a variant
   :kbd:`home`                            ,              , :kbd:`✔`, goto initial move
   :kbd:`end`                              ,              , :kbd:`✔`, goto end-of-game
   :kbd:`mouse-left-press`            , :kbd:`✔`,              , begin move
   :kbd:`mouse-left-release`          , :kbd:`✔`,              , end move
   :kbd:`Control-P`                      ,              , :kbd:`✔`, promote variant
   :kbd:`Control-D`                      ,              , :kbd:`✔`, demote variant
   :kbd:`Control-R`                      ,              , :kbd:`✔`, remove variant
   :kbd:`Control-M`                      ,              , :kbd:`✔`, promote variant to mainline
   :kbd:`Control-W`                      ,              , :kbd:`✔`, toggle warn of danger

In addition, several *standard* key strokes are supported:

.. csv-table:: Other Key-Codes
   :header: "Key", "Description"
   :widths: 30, 50

   :kbd:`Control-Z`                       , undo last move
   :kbd:`Control-F`                       , flip board
   :kbd:`Control-C`                       , copy game (PGN) to clipboard
   :kbd:`Control-V`                       , paste game (PGN) from clipboard
   :kbd:`Control-Shift-C`               , copy position (FEN) to clipboard
   :kbd:`Control-Shift-V`               , paste position (FEN) from clipboard

.. |MainWindow| image:: mainWindow.png
  :width: 800
  :alt: Main Window
.. _PGN: https://github.com/fsmosca/PGN-Standard
.. _UCI: http://wbec-ridderkerk.nl/html/UCIProtocol.html
'''

from typing import Optional, Callable, Dict, List, Tuple, Union, Any
import configparser
import os, os.path
import pickle
import io
import re
import sys

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import MzChess

if MzChess.useQt5:
 from PyQt5 import QtWidgets, QtGui, QtCore
 from PyQt5.QtWidgets import QAction
 from PyQt5 import uic
else:
 from PyQt6 import QtWidgets, QtGui, QtCore
 from PyQt6.QtGui import QAction
 from PyQt6 import uic

import chess, chess.pgn
from MzChess import read_game

import AboutDialog

class ChessMainWindow(QtWidgets.QMainWindow):
 logSignal = QtCore.pyqtSignal(str)
 notifySignal = QtCore.pyqtSignal(str)
 notifyGameSelectedSignal = QtCore.pyqtSignal(int)
 notifyGameHeadersChangedSignal = QtCore.pyqtSignal(chess.pgn.Headers)

 notifyGameNodeSelectedSignal = QtCore.pyqtSignal(chess.pgn.GameNode)
 notifyGameChangedSignal = QtCore.pyqtSignal(chess.pgn.Game)
 notifyNewGameNodeSignal = QtCore.pyqtSignal(chess.pgn.GameNode)
 fileDialogOptions = QtWidgets.QFileDialog.Option.DontUseNativeDialog
 fileDirectory = os.path.dirname(os.path.abspath(__file__))
 intRe = re.compile(r"^[+-]?[1-9][0-9]*$")
 boolRe = re.compile(r"^(True|False)$")

 def __init__(self, parent = None) -> None:
  super(ChessMainWindow, self).__init__(parent)
  MzChess.installLeipFont()
  uic.loadUi(os.path.join(self.fileDirectory, 'chessMainWindow.ui'), self)
  # self.setupUi(self)

  icon = QtGui.QIcon()
  icon.addPixmap(QtGui.QPixmap(os.path.join(self.fileDirectory,'schach.png')), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
  self.setWindowIcon(icon)
  
  self.pgm = MzChess.__name__
  self.version = MzChess.__version__
  self.dateString = MzChess.__date__
  
  # self.helpIndex = QtCore.QUrl.fromLocalFile(os.path.join(os.path.dirname(self.fileDirectory), 'doc_build', 'html', 'index.html'))
  self.helpIndex = QtCore.QUrl('https://reinhardm-dev.github.io/MzChess')

  self.ecoDB = MzChess.ECODatabase()
  self.ecoFen2IdDict = self.ecoDB.fen2Id()
  self.toolBar = QtWidgets.QToolBar()
  self._setIcon(self.actionNewDB, QtWidgets.QStyle.StandardPixmap.SP_FileDialogNewFolder)
  self._setIcon(self.actionOpenDB, QtWidgets.QStyle.StandardPixmap.SP_DirOpenIcon)
  self._setIcon(self.actionSaveDB, QtWidgets.QStyle.StandardPixmap.SP_DirHomeIcon)
  self._setIcon(self.actionSaveGame, QtWidgets.QStyle.StandardPixmap.SP_DialogSaveButton)
  self.toolBar.addSeparator()
  self._setIcon(self.actionNewGame, QtWidgets.QStyle.StandardPixmap.SP_FileIcon)
  self._setIcon(self.actionFlipBoard, QtWidgets.QStyle.StandardPixmap.SP_BrowserReload)
  self.toolBar.addSeparator()
  self._setIcon(self.actionNextMove, QtWidgets.QStyle.StandardPixmap.SP_ArrowDown)
  self._setIcon(self.actionPreviousMove, QtWidgets.QStyle.StandardPixmap.SP_ArrowUp)
  self._setIcon(self.actionUndoLastMove, QtWidgets.QStyle.StandardPixmap.SP_BrowserStop)
  self._setIcon(self.actionNextVariant, QtWidgets.QStyle.StandardPixmap.SP_ArrowForward)
  self._setIcon(self.actionPreviousVariant, QtWidgets.QStyle.StandardPixmap.SP_ArrowBack)
  self._setIcon(self.actionPromoteVariant, QtWidgets.QStyle.StandardPixmap.SP_MediaSeekBackward)
  self._setIcon(self.actionDemoteVariant, QtWidgets.QStyle.StandardPixmap.SP_MediaSeekForward)
  self._setIcon(self.actionPromoteVariant2Main, QtWidgets.QStyle.StandardPixmap.SP_MediaSkipBackward)
  self._setIcon(self.actionDeleteVariant, QtWidgets.QStyle.StandardPixmap.SP_DialogCloseButton)
  self.addToolBar(self.toolBar)
  
  self.gameOptions = { 
   'showOptions' : (self.actionShowOptions, False), 
   'warnOfDanger' : (self.actionWarnOfDanger, False), 
   'encoding' : (self.menuEncoding, 'UTF-8')
  }
  self.encodingDict = { 'UTF-8' : 'utf-8-sig', 'ISO 8859/1' : 'iso-8859-1',  'ASCII' : 'ascii'}
  
  self.engineOptions = {
   'selectedEngine' : (self.menuSelectEngine, None), 
   'searchDepth' : (self.menuSearchDepth, 15), 
   'blunderLimit' : (self.menuBlunderLimit, None), 
   'numberOfAnnotations' : (self.menuNumberOfAnnotations, 1), 
   'annotateVariants' : (self.menuAnnotateVariants, None), 
   'showScores' : (self.actionShowScores, False), 
   'showHints' :( self.actionShowHints, False) 
   }
  
  self.engineDict = dict()
  self.recentPGN = dict()
  self.eventList = list()
  self.siteList = list()
  self.playerList = list()
  self.optionalHeaderItems = list()

  self.hintEngine = None
  self.mateScore = 100
  self.debugEngine = False
  
  self.settingsFile = os.path.join(self.fileDirectory, 'settings.ini')
  self.settings = configparser.ConfigParser(delimiters=['='], allow_no_value=True)
  self.settings.optionxform = str
  if os.path.isfile(self.settingsFile):
   self.settings.read(self.settingsFile, encoding = 'utf-8')
   self.engineDict = MzChess.loadEngineSettings(self.settings)
   if 'Recent' in self.settings.sections():
    self.recentPGN = dict()
    for recentDB, enc in dict(self.settings['Recent']).items():
     if os.path.exists(recentDB):
      self.recentPGN[recentDB] = enc
   if 'Events' in self.settings.sections():
    self.eventList = self.settings.options('Events')
   if 'Sites' in self.settings.sections():
    self.siteList = self.settings.options('Sites')
   if 'Players' in self.settings.sections():
    self.playerList = self.settings.options('Players')
   if 'OptionalHeaderItems' in self.settings.sections():
    self.optionalHeaderItems = list(self.settings['OptionalHeaderItems'])
   if self.settings['Menu/Engine']['selectedEngine'] not in self.engineDict:
    self.settings['Menu/Engine']['selectedEngine'] = None
    self.settings['Menu/Engine']['showHints'] = False
    self.settings['Menu/Engine']['showScores'] = False

  self.menuRecentDB.clear()
  if self.recentPGN is None:
   self.recentPGN = dict()
  else:
   actDir = os.path.expanduser('~')
   for rItem in self.recentPGN.keys():
    if rItem is not None and os.path.isfile(rItem):
     if actDir == os.path.expanduser('~'):
      actDir = os.path.dirname(rItem)
     self.menuRecentDB.addAction(rItem)
  
  os.chdir(actDir)
  
  self.aboutDialog = AboutDialog.AboutDialog()
  self.aboutDialog.setup(
   pgm = self.pgm, 
   version = self.version, 
   dateString = self.dateString)

  sbText = self.statusBar().font()
  sbText.setPointSize(12)
  
  self.infoLabel = QtWidgets.QLabel()
  self.infoLabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignLeft)
  self.infoLabel.setFont(sbText)
  self.statusBar().addPermanentWidget(self.infoLabel, 150)
  self.engineLabel = QtWidgets.QLabel()
  self.engineLabel.setFont(sbText)
  self.engineLabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
  self.engineLabel.setToolTip("Engine in use")
  self.statusBar().addPermanentWidget(self.engineLabel, 50)
  self.hintLabel = QtWidgets.QLabel()
  self.hintLabel.setFont(sbText)
  self.hintLabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
  self.hintLabel.setToolTip("Best move/Score[Pawns]")
  self.statusBar().addPermanentWidget(self.hintLabel, 30)
  self.ecoLabel = QtWidgets.QLabel()
  self.ecoLabel.setFont(sbText)
  self.ecoLabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
  self.statusBar().addPermanentWidget(self.ecoLabel, 10)
  self.squareLabel = QtWidgets.QLabel()
  self.squareLabel.setFont(sbText)
  self.squareLabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
  self.squareLabel.setToolTip("Position")
  self.statusBar().addPermanentWidget(self.squareLabel, 10)

  self.itemSelector = MzChess.ItemSelector('Header Elements (without 7-tag roster)...', pointSize = 10)
  
  self.loadOptionDict('Menu/Game', self.gameOptions)
  self.resetSelectEngine()
  self.loadOptionDict('Menu/Engine', self.engineOptions)
  self.gameListTableView.setup(self.notifyGameSelectedSignal)
  self.notifyGameSelectedSignal.connect(self.gameSelected)
  self.gameListFile = ''
  self.gameList = list()
  self.gameListChanged = False
  self.gameFile = ''
  self.game = chess.pgn.Game()
  self.gameNode = self.game
  self.gameID = None
  self._gameChanged = False

  self.gameHeaderTableView.setup(
       notifyGameHeadersChangedSignal = self.notifyGameHeadersChangedSignal, 
       eventList = self.eventList, siteList = self.siteList, playerList = self.playerList)
  self.notifyGameHeadersChangedSignal.connect(self.gameHeadersChanged)

  self.boardGraphicsView.setup(
   notifyNewGameNodeSignal = self.notifyNewGameNodeSignal, notifyGameNodeSelectedSignal =self.notifyGameNodeSelectedSignal, 
   materialLabel = self.materialLabel, squareLabel = self.squareLabel, 
   turnFrame= self.turnFrame, hintLabel = self.hintLabel, 
   flipped = False)
  self.gameTreeViewWidget.setup(self.notifyGameNodeSelectedSignal, self.notifyGameChangedSignal)
  self.scorePlotGraphicsView.setup(self.notifyGameNodeSelectedSignal)
  self.notifyGameNodeSelectedSignal.connect(self.gameNodeSelected)
  self.notifyNewGameNodeSignal.connect(self.newGameNode)
  self.notifyGameChangedSignal.connect(self.gameChanged)
  self.notifySignal.connect(self.notify)
  self.logSignal.connect(self.toLog)

 def setup(self) -> None:
  sizes = self.splitter.sizes()
  self.splitter.setSizes([self.boardGraphicsView.height(), sizes[0] + sizes[1] - self.boardGraphicsView.height()])
  self.boardGraphicsView.setDrawOptions(self.actionShowOptions.isChecked())
  self.boardGraphicsView.setGameNode(self.game)
  self.gameTreeViewWidget.setGame(self.game)
  self.gameListTableView.resetDB()
  self.gameHeaderTableView.resetGame()

 def loadOptionDict(self, name : str, n2oDict : Dict[str, Union[QAction, Tuple[QtWidgets.QMenu, Any]]]) -> Dict[str, Union[int, float]]:
  if name not in self.settings:
   self.settings[name] = dict()
  optionsDict = dict()
  for opt, o2v in n2oDict.items():
   obj, defValue = o2v
   optionsDict[opt] = defValue
   if opt in self.settings[name]:
    optValue = str(self.settings[name][opt])
   else:
    optValue = str(defValue)
   if isinstance(obj, QAction):
    optionsDict[opt] = (optValue == 'True')
    obj.setChecked(optionsDict[opt])
   elif isinstance(obj, QtWidgets.QMenu):
    for action in obj.actions():
     tListString = action.text()
     if len(tListString) == 0:
      continue
     tList = tListString.split(' ')
     if self.intRe.match(tList[0]) is not None:
      actionValue = int(tList[0])
     elif tList[0] == 'All':
      actionValue = 2^32 - 1
     elif tList[0] == 'None':
      actionValue = None
     else:
      try:
       actionValue = float(tList[0])
      except:
       actionValue = tListString
     if str(actionValue) == optValue or '{}.0'.format(actionValue) == optValue:
      action.setChecked(True)
      optionsDict[opt] = actionValue
     else:
      action.setChecked(False)
  self.settings[name] = optionsDict

 def resetSelectEngine(self) -> None:
  self.menuSelectEngine.clear()
  action = self.menuSelectEngine.addAction('None')
  action.setCheckable(True)
  if 'Menu/Engine' in self.settings.sections():
   if self.settings['Menu/Engine']['selectedEngine'] is None or self.settings['Menu/Engine']['selectedEngine'] not in self.engineDict:
    action.setChecked(True)
   self.menuSelectEngine.addSeparator()
   for eItem in self.engineDict:
    action = self.menuSelectEngine.addAction(eItem)
    action.setCheckable(True)
    if self.settings['Menu/Engine']['selectedEngine'] == eItem:
     action.setChecked(True)

 def updateSettingsList(self, section : str, valueList : List[Union[str, Tuple[str, str]]] = list(), firstValue : Union[str, Tuple[str, str], None] = None) -> None:
  if section not in self.settings.sections():
   self.settings.add_section(section)
  newDict = dict()
  if len(valueList) > 0:
   newDict = dict(self.settings[section])
  if isinstance(firstValue, tuple):
   key = firstValue[0]
   value = firstValue[1]
  else:
   key = firstValue
   value = None
  keys = list()
  if key is not None:
   newDict[key] = value
   self.settings.remove_option(section, key)
   keys = [key]
  if len(valueList) == 0:
   keys += self.settings.options(section)
  self.settings.remove_section(section)
  self.settings.add_section(section)
  for key in keys:
   if key in newDict:
    value = newDict[key]
   self.settings[section][key] = value
  for kv in valueList:
   if isinstance(kv, tuple):
    self.settings[section][kv[0]] = kv[1]
   else:
    self.settings[section][kv] = None
  
 def saveSettings(self) -> None:
  with open(self.settingsFile, 'w', encoding = 'utf-8') as f:
   self.settings.write(f)

 # --------------------------------------------------------------------------------------------------
 
 @QtCore.pyqtSlot(str)
 def notifyError(self, str : str) -> None:
  msgBox = QtWidgets.QMessageBox()
  msgBox.setIcon(QtWidgets.QMessageBox.Icon.Critical)
  msgBox.setText(str)
  msgBox.setWindowTitle("Error ...")
  msgBox.exec()

 def notify(self, str):
  self.infoLabel.setText(str)
  self.infoLabel.update()
  QtWidgets.QApplication.processEvents()

 @QtCore.pyqtSlot(str)
 def toLog(self, msg):
  self.uciTextEdit.moveCursor (QtGui.QTextCursor.MoveOperation.End)
  self.uciTextEdit.insertPlainText (msg)
  self.uciTextEdit.moveCursor(QtGui.QTextCursor.MoveOperation.End)

 def _setIcon(self, action, stdIcon, toToolbar = True):
  action.setIcon(self.style().standardIcon(stdIcon))
  if toToolbar:
   self.toolBar.addAction(action)

 def _showEcoCode(self, gameNode, fromBeginning = False):
  if fromBeginning:
   actNode = gameNode.game().next()
   self.ecoCode = '---'
   self.ecoDescription = ''
  else:
   actNode = gameNode
  id = -1
  while actNode is not None:
   fen = ' '.join(actNode.board().fen().split(' ')[:-2])
   if fen in self.ecoFen2IdDict:
    id = self.ecoFen2IdDict[fen]
   actNode = actNode.next()
  if id >= 0:
   self.ecoCode = self.ecoDB[id][0]
   self.ecoDescription = self.ecoDB[id][1]
   gameNode.game().headers['Opening'] = self.ecoDescription
   gameNode.game().headers['ECO'] = self.ecoCode
   self.ecoLabel.setText(self.ecoCode)
   self.ecoLabel.setToolTip(self.ecoDescription)

 def _allowNewGameList(self):
  self.notify('')
  if self.gameListChanged:
   rc = QtWidgets.QMessageBox.warning(self, 
    'Game database not saved ...', 'Do you like to close anyway ?', 
     QtWidgets.QMessageBox.StandardButton.Yes | QtWidgets.QMessageBox.StandardButton.No, 
     QtWidgets.QMessageBox.StandardButton.No)
   if rc == QtWidgets.QMessageBox.StandardButton.No:
    return False
  self.gameListTableView.resetDB()
  self.gameHeaderTableView.resetGame()
  self.gameListChanged = False
  self.gameList = list()
  self.gameID = None
  return True

 def _allowNewGame(self):
  self.notify('')
  if self._gameChanged:
   rc = QtWidgets.QMessageBox.warning(self, 
    'Game not saved ...', 'Do you like to close anyway ?', 
     QtWidgets.QMessageBox.StandardButton.Yes | QtWidgets.QMessageBox.StandardButton.No, 
     QtWidgets.QMessageBox.StandardButton.No)
   if rc == QtWidgets.QMessageBox.StandardButton.No:
    return False
  self.game = chess.pgn.Game()
  self.game.headers = self._setGameHeader(self.optionalHeaderItems)
  self.gameNode = self.game
  self.boardGraphicsView.setGameNode(self.game)
  self.gameTreeViewWidget.setGame(self.game)
  self.gameHeaderTableView.resetGame()
  self._showEcoCode(self.game, fromBeginning = True)
  self.setChessWindowTitle()
  self._gameChanged = False
  return True
 
 def _setGameHeader(self, optionalHeaderItems : List[str]) -> chess.pgn.Headers:
  selectedItems =  list(chess.pgn.Headers()) + optionalHeaderItems
  newHeaders = chess.pgn.Headers()
  for el in selectedItems:
   if el in self.game.headers:
    newHeaders[el] = self.game.headers[el]
   else:
    newHeaders[el] = self.gameHeaderTableView.stdKey2ValueTypeDict[el][0]
  self.optionalHeaderItems = optionalHeaderItems
  return newHeaders

 # ---------------------------------------------------------------------------

 def openPGN(self, pgnFile : str, encoding : str):
  if not self._allowNewGameList():
   return
  self.notify('Opening {} ...'.format(os.path.basename(pgnFile)))
  self.gameListFile = os.path.split(pgnFile)[1]
  _,  ext = os.path.splitext(pgnFile)
  if ext == '.ppgn':
   try:
    with open(pgnFile, mode = 'rb') as f:
     self.gameList += pickle.load(f)
     encoding = None
   except:
    self.notifyError('Cannot open PPGN file {}'.format(pgnFile))
    return
  else:
   try:
    pgn = open(pgnFile, mode = 'r', encoding = encoding)
   except:
    self.notifyError('Cannot open PGN file {}'.format(pgnFile))
    return
   n = 1
   self.gameList = list()
   while True:
    actGame = read_game(pgn)
    if actGame is None:
     break
    if len(actGame.errors) == 0:
     self.gameList.append(actGame)
    else:
     self.notifyError('Failed to load game #{}'.format(n))
    n += 1
  os.chdir(os.path.dirname(pgnFile))
  self.gameListTableView.setGameList(self.gameList)
  self.settings['Recent'] = {pgnFile : encoding}
  newRecentPGN = {pgnFile : encoding}
  for file, enc in self.recentPGN.items():
   if file != pgnFile:
    newRecentPGN[file] = enc
    self.settings['Recent'][file] = enc
  self.recentPGN = newRecentPGN
  self.saveSettings()
  self.menuRecentDB.clear()
  for rItem in self.recentPGN:
   if rItem is not None and os.path.isfile(rItem):
    self.menuRecentDB.addAction(rItem)
  os.chdir(os.path.dirname(pgnFile))
  self.notify('')
 
 @QtCore.pyqtSlot(QAction)
 def on_menuEncoding_triggered(self, action):
  self.notify('')
  for actAction in self.menuEncoding.actions():
   actAction.setChecked(False)
  action.setChecked(True)
  self.settings['Menu/Game']['encoding'] = action.text()
  self.saveSettings()

 @QtCore.pyqtSlot()
 def on_actionNewDB_triggered(self):
  self.notify('')
  self.on_actionCloseDB_triggered()

 @QtCore.pyqtSlot()
 def on_actionOpenDB_triggered(self):
  self.notify('')
  pgnFile, _ = QtWidgets.QFileDialog.getOpenFileName(self,"Load Game Database ...", "",
        "Pickle Portable Game Notation Files (*.ppgn);;Portable Game Notation Files (*.pgn);;All Files (*)", options = self.fileDialogOptions)
  if pgnFile is not None and len(pgnFile) > 0:
   self.openPGN(pgnFile, self.encodingDict[self.settings['Menu/Game']['encoding']])
   
 @QtCore.pyqtSlot(QAction)
 def on_menuRecentDB_triggered(self, action):
  self.notify('')
  fileName = action.text()
  self.openPGN(fileName, self.recentPGN[fileName])
 
 @QtCore.pyqtSlot()
 def on_actionCloseDB_triggered(self):
  self.notify('')
  if not self._allowNewGameList() or not self._allowNewGame():
   return
  self.gameListTableView.resetDB()
  self.gameHeaderTableView.resetGame()
  self.gameList = list()
  self.gameID = None

 @QtCore.pyqtSlot(QtGui.QCloseEvent)
 def closeEvent(self, ev):
  self.notify('')
  if not self._allowNewGameList() or not self._allowNewGame():
   ev.ignore()
  else:
   if self.hintEngine is not None:
    self.hintEngine.kill(True)
   ev.accept()

 @QtCore.pyqtSlot()
 def on_actionExit_triggered(self):
  self.notify('')
  if self._allowNewGameList() and self._allowNewGame():
   self.close()

 def saveDB(self, forceAppend):
  if forceAppend:
   mode = 'a'
  else:
   mode = 'w'
  self.notify('')
  if len(self.gameList) == 0:
    self.notifyError('No Game Database available')
    return
  pgnFile, _ = QtWidgets.QFileDialog.getSaveFileName(self,
     "Save Game Database ...", self.gameListFile,
     "Pickle Portable Game Notation Files (*.ppgn);;Portable Game Notation Files (*.pgn);;All Files (*)", options = self.fileDialogOptions)
  if pgnFile is None or len(pgnFile) == 0:
   return
  _,  ext = os.path.splitext(pgnFile)
  if ext == '.ppgn':
   try:
    with open(pgnFile, mode = '{}b'.format(mode)) as f:
     pickle.dump(self.gameList, f)
   except:
    self.notifyError('Cannot open PPGN file {}'.format(pgnFile))
    return
   encoding = None
  elif ext == '.pgn':
   try:
    encoding = self.settings['Menu/Game']['encoding']
    with open(pgnFile, mode = mode,  encoding = self.encodingDict[encoding]) as f:
     pgnString = str()
     for game in self.gameList:
      exporter = chess.pgn.StringExporter(headers=True, variations=True, comments=True)
      pgnString += game.accept(exporter)
      if game != self.gameList[-1]:
       pgnString += '\n\n'
     f.write(pgnString)
   except:
    self.notifyError('Cannot save PGN file {}'.format(pgnFile))
    return
  else:
    self.notifyError('PGN file {} has improper extension (.pgn or .pgn expected)'.format(pgnFile))
  self.updateSettingsList('Recent', self.recentPGN.items(), firstValue = (pgnFile, encoding))
  self.saveSettings()
  self.gameListChanged = False
  self.gameListFile = os.path.split(pgnFile)[1]
  os.chdir(os.path.dirname(pgnFile))

 @QtCore.pyqtSlot()
 def on_actionAppend2DB_triggered(self):
  self.saveDB(True)
  
 @QtCore.pyqtSlot()
 def on_actionSaveDB_triggered(self):
  self.saveDB(False)
   
 @QtCore.pyqtSlot()
 def on_actionAddGame_triggered(self):
  self.notify('')
  if self.gameID is None:
   self.gameID = len(self.gameList)
   self.gameList.append(self.game)
   self.gameListTableView.setGameList(self.gameList)
  else:
   self.gameList[self.gameID] = self.game
  self.gameListChanged = True
   
 @QtCore.pyqtSlot()
 def on_actionRemoveGames_triggered(self):
  self.notify('')
  selectedIndices = self.gameListTableView.selectedIndexes()
  if len(selectedIndices) != 0:
   removeList = list()
   for index in selectedIndices:
    row = index.row()
    removeList.append(row)
    if row == self.gameID:
     self.gameID = None
   newGameList = list()
   for n, game in enumerate(self.gameList):
    if n not in removeList:
     newGameList.append(game)
   self.gameList = newGameList
   self.gameListChanged = True
   self.gameListTableView.setGameList(self.gameList)
  
 @QtCore.pyqtSlot(bool)
 def on_actionNewGame_triggered(self):
  self._allowNewGame()
  
 @QtCore.pyqtSlot(QAction)
 def on_menuEndGame_triggered(self, action):
  self.notify('')
  endGame = self.game.end()
  if endGame.board().is_game_over():
   self.notifyError('You cannot change the game result any more')
   return
  self.game.headers['Result'] = action.text()
  self.boardGraphicsView.setGameNode(self.gameNode)
  self.gameChanged(self.game)
  
 @QtCore.pyqtSlot()
 def on_actionSaveGame_triggered(self):
  self.notify('')
  pgnFile, _ = QtWidgets.QFileDialog.getSaveFileName(self,
     "Save Game ...", self.gameFile,"Portable Game Notation Files (*.pgn);;All Files (*)", options = self.fileDialogOptions)
  if pgnFile is not None and len(pgnFile) > 0:
   try:
    encoding = self.settings['Menu/Game']['encoding']
    f = open(pgnFile, mode = 'w',  encoding = self.encodingDict[encoding])
   except:
    self.notifyError('Cannot open PGN file {}'.format(pgnFile))
    return
   exporter = chess.pgn.StringExporter(headers=True, variations=True, comments=True)
   pgnString = self.game.accept(exporter)
   f.write(pgnString)
   f.close()
   self.updateSettingsList('Recent', self.recentPGN.items(), firstValue = (pgnFile, encoding))
   self.saveSettings()
   self._gameChanged = False
   self.gameFile = os.path.split(pgnFile)[1]
   os.chdir(os.path.dirname(pgnFile))

 # ---------------------------------------------------------------------------

 @QtCore.pyqtSlot(QtGui.QWheelEvent)
 def wheelEvent(self, ev):
  self.notify('')
  numDegrees = ev.angleDelta()
  if numDegrees.y() < 0:
   self.on_actionNext_Move_triggered()
  elif numDegrees.y() > 0:
   self.on_actionPrevious_Move_triggered()
  ev.accept()

 @QtCore.pyqtSlot()
 def on_actionNextMove_triggered(self):
  self.notify('')
  self.gameNode = self.boardGraphicsView.nextMove()
  self.gameTreeViewWidget.selectNodeItem(self.gameNode)
  self.scorePlotGraphicsView.selectNodeItem(self.gameNode)

 @QtCore.pyqtSlot()
 def on_actionPreviousMove_triggered(self):
  self.notify('')
  self.gameNode = self.boardGraphicsView.previousMove()
  self.gameTreeViewWidget.selectNodeItem(self.gameNode)
  self.scorePlotGraphicsView.selectNodeItem(self.gameNode)

 @QtCore.pyqtSlot()
 def on_actionNextVariant_triggered(self):
  self.notify('')
  self.gameTreeViewWidget.selectSubnodeItem(self.gameNode, next = True)

 @QtCore.pyqtSlot()
 def on_actionPreviousVariant_triggered(self):
  self.notify('')
  self.gameTreeViewWidget.selectSubnodeItem(self.gameNode, next = False)

 @QtCore.pyqtSlot()
 def on_actionPromoteVariant_triggered(self):
  self.notify('')
  if self.gameNode.is_main_variation():
   self.notifyError('You cannot promote from main line nodes')
   return
  self.gameTreeViewWidget.promoteVariant()

 @QtCore.pyqtSlot()
 def on_actionPromoteVariant2Main_triggered(self):
  self.notify('')
  if self.gameNode.is_main_variation():
   self.notifyError('You cannot promote from main line nodes')
   return
  self.gameTreeViewWidget.promoteVariant2Main()

 @QtCore.pyqtSlot()
 def on_actionDemoteVariant_triggered(self):
  self.notify('')
  if self.gameNode.is_main_variation():
   self.notifyError('You cannot promote from main line nodes')
   return
  self.gameTreeViewWidget.demoteVariant()

 @QtCore.pyqtSlot()
 def on_actionDeleteVariant_triggered(self):
  self.notify('')
  if self.gameNode.parent is None:
   self.notifyError('Use Game/New to remove a complete game')
   return
  self.gameTreeViewWidget.removeVariant()
  self.gameNode = self.gameNode.parent
  self.boardGraphicsView.setGameNode(self.gameNode)

 @QtCore.pyqtSlot()
 def on_actionUndoLastMove_triggered(self):
  self.notify('')
  if self.gameNode is None or self.gameNode.parent is None:
   return
  if not self.gameNode.is_end():
   self.notifyError('Select the end of the mainline or a variant')
   return
  gameNode = self.gameNode
  self.gameNode = self.gameNode.parent
  self.boardGraphicsView.setGameNode(self.gameNode)
  self.gameTreeViewWidget.removeGameNode(gameNode)
  if gameNode.is_mainline():
   self.scorePlotGraphicsView.removeLastNode()
  self.gameNode.remove_variation(gameNode)
  self._gameChanged = True

 def setChessWindowTitle(self):
  self.setWindowTitle('MzChess {}: {} - {}'.format(self.game.headers['Date'], self.game.headers['White'], self.game.headers['Black']))

 # ---------------------------------------------------------------------------

 @QtCore.pyqtSlot(QAction)
 def on_menuSelectEngine_triggered(self, action):
  self.notify('')
  oldValue = self.settings['Menu/Engine']['selectedEngine']
  for actAction in self.menuSelectEngine.actions():
   actAction.setChecked(False)
  action.setChecked(True)
  self.settings['Menu/Engine']['selectedEngine'] = action.text()
  if oldValue is None:
   self.show_HintsScores()
  self.saveSettings()
  
 @QtCore.pyqtSlot(QAction)
 def on_menuSearchDepth_triggered(self, action):
  self.notify('')
  for actAction in self.menuSearchDepth.actions():
   actAction.setChecked(False)
  action.setChecked(True)
  tList = action.text().split(' ')
  self.settings['Menu/Engine']['searchDepth'] = str(int(tList[0]))
  self.saveSettings()
  
 @QtCore.pyqtSlot(QAction)
 def on_menuNumberOfAnnotations_triggered(self, action):
  self.notify('')
  for actAction in self.menuBlunderLimit.actions():
   actAction.setChecked(False)
  action.setChecked(True)
  naValue = action.text()
  self.settings['Menu/Engine']['numberOfAnnotations'] = naValue
  self.saveSettings()

 @QtCore.pyqtSlot(QAction)
 def on_menuBlunderLimit_triggered(self, action):
  self.notify('')
  for actAction in self.menuBlunderLimit.actions():
   actAction.setChecked(False)
  action.setChecked(True)
  blValue = action.text()
  if blValue == 'None':
   blunderLimit = None
  else:
   blunderLimit = -float(blValue.split(' ')[0])
  self.settings['Menu/Engine']['blunderLimit'] = str(blunderLimit)
  self.saveSettings()
   
 @QtCore.pyqtSlot(QAction)
 def on_menuAnnotateVariants_triggered(self, action):
  self.notify('')
  for actAction in self.menuAnnotateVariants.actions():
   actAction.setChecked(False)
  action.setChecked(True)
  avValue = action.text().split(' ')
  if avValue == 'None':
   self.settings['Menu/Engine']['annotateVariants'] = str(None)
  elif avValue == 'All':
   self.settings['Menu/Engine']['annotateVariants'] = str(2^32 - 1)
  else:
   self.settings['Menu/Engine']['annotateVariants'] = str(int(avValue[0]))
  self.saveSettings()

 @QtCore.pyqtSlot()
 def on_actionAnnotateLastMove_triggered(self):
  if self.settings['Menu/Engine']['selectedEngine'] is None:
   self.notifyError('No engine selected')
   return
  self.notify('')
  if self.debugEngine:
   logFunction = self.logSignal.emit
  else:
   logFunction = None
  engine = MzChess.ChessEngine(
      self.engineDict[self.settings['Menu/Engine']['selectedEngine']], 
      limit = chess.engine.Limit(depth = self.settings['Menu/Engine']['searchDepth']), 
      log = logFunction)

  aEngine = MzChess.AnnotateEngine(notifyFunction = self.notifySignal.emit)
  annotateVariants = self.settings['Menu/Engine']['annotateVariants']
  if annotateVariants is not None and annotateVariants.isdigit():
   annotateVariants = int(annotateVariants)
  else:
   annotateVariants = 0
  aEngine.setup(engine, hintPLYs = annotateVariants, multiPV = int(self.settings['Menu/Engine']['numberOfAnnotations']))
  if aEngine.run(self.gameNode.parent, numberOfPlys = 2):
   annotator = MzChess.Annotator(self.settings['Menu/Engine']['selectedEngine'], notifyFunction = self.notifySignal.emit)
   if self.settings['Menu/Engine']['blunderLimit'] is None:
    annotator.setBlunder(-float('inf'), addVariant = False)
   else:
    annotator.setBlunder(float(self.settings['Menu/Engine']['blunderLimit']), addVariant = True)
   annotator.apply(game = self.gameNode, scoreListList = aEngine.scoreListList, pvListList = aEngine.pvListList, forceHints = True)
   self.scorePlotGraphicsView.setGame(self.game)
   self.gameTreeViewWidget.setGame(self.game)
   self.gameHeaderTableView.setGame(self.game)
   self._gameChanged = True
  else:
   self.notifyError('Annotation failed')

 @QtCore.pyqtSlot()
 def on_actionAnnotateAll_triggered(self):
  if self.settings['Menu/Engine']['selectedEngine'] is None:
   self.notifyError('No engine selected')
   return
  self.notify('')
  if self.debugEngine:
   logFunction = self.logSignal.emit
  else:
   logFunction = None
  engine = MzChess.ChessEngine(
      self.engineDict[self.settings['Menu/Engine']['selectedEngine']], 
      limit = chess.engine.Limit(depth = self.settings['Menu/Engine']['searchDepth']), 
      log = logFunction)

  aEngine = MzChess.AnnotateEngine(notifyFunction = self.notifySignal.emit)
  annotateVariants = self.settings['Menu/Engine']['annotateVariants']
  if annotateVariants is not None and annotateVariants.isdigit():
   annotateVariants = int(annotateVariants)
  else:
   annotateVariants = 0
  aEngine.setup(engine, hintPLYs = annotateVariants, multiPV = int(self.settings['Menu/Engine']['numberOfAnnotations']))
  if aEngine.run(self.game, numberOfPlys = None):
   annotator = MzChess.Annotator(self.settings['Menu/Engine']['selectedEngine'])
   if self.settings['Menu/Engine']['blunderLimit'] is None:
    annotator.setBlunder(-float('inf'), addVariant = False)
   else:
    annotator.setBlunder(float(self.settings['Menu/Engine']['blunderLimit']), addVariant = True)
   self.game = annotator.apply(game = self.game, scoreListList = aEngine.scoreListList, pvListList = aEngine.pvListList)
   self.scorePlotGraphicsView.setGame(self.game)
   self.gameTreeViewWidget.setGame(self.game)
   self.gameHeaderTableView.setGame(self.game)
   self._gameChanged = True
  else:
   self.notifyError('Annotation failed')

 @QtCore.pyqtSlot(bool)
 def on_actionShowOptions_toggled(self, checked):
  self.notify('')
  self.boardGraphicsView.setDrawOptions(checked)
  self.settings['Menu/Game']['showOptions'] = str(checked)
  self.saveSettings()

 @QtCore.pyqtSlot(bool)
 def on_actionWarnOfDanger_toggled(self, checked):
  self.notify('')
  self.boardGraphicsView.setWarnOfDanger(checked)
  self.settings['Menu/Game']['warnOfDanger'] = str(checked)
  self.saveSettings()

 @QtCore.pyqtSlot(bool)
 def on_actionFlipBoard_toggled(self, checked):
  self.notify('')
  self.boardGraphicsView.setFlipped(checked)
  
 @QtCore.pyqtSlot()
 def on_actionSelectHeaderElements_triggered(self):
  self.notify('')
  headerElements = self.gameHeaderTableView.headerElements(withoutSevenTagRoster = True)
  self.itemSelector.setContent(headerElements, self.optionalHeaderItems)
  if not self.itemSelector.exec():
   return None
  self.game.headers = self._setGameHeader(self.itemSelector.selectedItems() )
  self.gameHeaderTableView.setGame(self.game)
  self.updateSettingsList('OptionalHeaderItems', self.optionalHeaderItems)
  self.saveSettings()
   
 def show_HintsScores(self):
  if self.settings['Menu/Engine']['selectedEngine'] is None:
   self.notifyError('No engine selected')
   return
  self.notify('')
  hintsChecked = self.actionShowHints.isChecked()
  scoresChecked = self.actionShowScores.isChecked()
  self.settings['Menu/Engine']['showHints'] = str(hintsChecked)
  self.settings['Menu/Engine']['showScores'] = str(scoresChecked)
  self.saveSettings()
  if hintsChecked or scoresChecked:
   if self.settings['Menu/Engine']['searchDepth'] is None:
    self.notifyError('"Engine/Search Depth" undefined.')
   if self.debugEngine:
    logFunction = self.logSignal.emit
   else:
    logFunction = None
   self.hintEngine = MzChess.ChessEngine(
      self.engineDict[self.settings['Menu/Engine']['selectedEngine']], 
      limit = chess.engine.Limit(depth = self.settings['Menu/Engine']['searchDepth']), 
      log = logFunction)
   self.boardGraphicsView.setHint(enableHint = hintsChecked, enableScore = scoresChecked, engine = self.hintEngine)
   self.engineLabel.setText(self.settings['Menu/Engine']['selectedEngine'])
  else:
   self.hintEngine = None
   self.boardGraphicsView.setHint(enableHint = False, enableScore = False, engine = None)
   self.engineLabel.setText('---')

 @QtCore.pyqtSlot(bool)
 def on_actionShowHints_toggled(self, checked):
  self.show_HintsScores()
    
 @QtCore.pyqtSlot(bool)
 def on_actionShowScores_toggled(self, checked):
  self.show_HintsScores()
  
 @QtCore.pyqtSlot()
 def on_actionConfigureEngine_triggered(self):
  self.notify('')
  configForm = MzChess.ConfigureEngine()
  newEngineDict = configForm.run(engineDict = self.engineDict, log = self.debugEngine)
  if newEngineDict is not None:
   self.engineDict = newEngineDict
   self.resetSelectEngine()
   MzChess.saveEngineSettings(self.settings, self.engineDict)
   self.saveSettings()
   
 @QtCore.pyqtSlot(bool)
 def on_actionDebugEngine_toggled(self, checked):
  self.notify('')
  self.debugEngine = checked
  if self.hintEngine is not None:
   if self.debugEngine:
    logFunction = self.logSignal.emit
   else:
    logFunction = None
   self.hintEngine.setLog(logFunction)
  
 @QtCore.pyqtSlot()
 def on_actionAbout_triggered(self):
  self.notify('')
  self.aboutDialog.exec()
 
 @QtCore.pyqtSlot()
 def on_actionHelp_triggered(self):
  self.notify('')
  QtGui.QDesktopServices.openUrl(self.helpIndex)

 # ---------------------------------------------------------------------------

 @QtCore.pyqtSlot() 
 def on_actionCopyFEN_triggered(self):
  self.notify('')
  fen = self.gameNode.board().fen()
  QtWidgets.QApplication.clipboard().setText(fen)

 @QtCore.pyqtSlot() 
 def on_actionPasteFEN_triggered(self):
  self.notify('')
  if not self._allowNewGame():
   return
  fen = QtWidgets.QApplication.clipboard().text()
  try:
   MzChess.checkFEN(fen)
  except ValueError as err:
   self.notifyError('Improper FEN {}:\n{}'.format(fen, str(err)))
   return
  self.gameID = None
  self.game = chess.pgn.Game()
  self.game.setup(fen)
  self.gameNode = self.game
  self._showEcoCode(self.game, fromBeginning = True)
  self.boardGraphicsView.setGameNode(self.game)
  self.gameTreeViewWidget.setGame(self.game)
  self.gameHeaderTableView.setGame(self.game)

 @QtCore.pyqtSlot() 
 def on_actionFENBuilder_triggered(self):
  self.fenBuilder = QtCore.QProcess()
  self.fenBuilder.start(sys.executable, [os.path.join(self.fileDirectory,'qbuildfen.py')], QtCore.QIODevice.NotOpen)
  
 @QtCore.pyqtSlot() 
 def on_actionCopyGame_triggered(self):
  self.notify('')
  exporter = chess.pgn.StringExporter(headers=True, variations=True, comments=True)
  pgnString = self.game.accept(exporter)
  QtWidgets.QApplication.clipboard().setText(pgnString)

 @QtCore.pyqtSlot() 
 def on_actionPasteGame_triggered(self):
  self.notify('')
  if not self._allowNewGame():
   return
  self.gameID = None
  pgnString = QtWidgets.QApplication.clipboard().text()
  pgn = io.StringIO(pgnString)
  self.game = read_game(pgn)
  self.gameNode = self.game
  if len(self.game.errors) != 0:
   self.notifyError('Improper PGN')
   self.game = chess.pgn.Game()
  self._showEcoCode(self.game, fromBeginning = True)
  self.boardGraphicsView.setGameNode(self.game)
  self.gameTreeViewWidget.setGame(self.game)
  self.gameHeaderTableView.setGame(self.game)
  self.scorePlotGraphicsView.setGame(self.game)
  
 # ---------------------------------------------------------------------------

 @QtCore.pyqtSlot(int)
 def gameSelected(self, gameID):
  if not self._allowNewGame():
   return
  self.gameID = gameID
  self.game = self.gameList[gameID]
  self.gameNode = self.game
  try:
   self.notify('Loading game #{} ...'.format(self.gameID))
   self._showEcoCode(self.game, fromBeginning = True)
   self.boardGraphicsView.setGameNode(self.game)
   self.gameTreeViewWidget.setGame(self.game)
   self.gameHeaderTableView.setGame(self.game)
   self.scorePlotGraphicsView.setGame(self.game)
   self.setChessWindowTitle()
   self.notify('')
   err = False
  except:
   err = True
  if err:
   self.notifyError('UIE: Improper game ="{}"'.format(self.game))

 @QtCore.pyqtSlot(chess.pgn.Headers)
 def gameHeadersChanged(self,  headers):
  self.notify('')
  if 'Comment' in headers:
   self.game.comment = headers['Comment']
   del headers['Comment']
  else:
   self.game.comment = ''
  if '?' not in headers["White"] and headers["White"] not in self.playerList:
   self.updateSettingsList('Players', self.playerList, firstValue = headers["White"])
  if '?' not in headers["Black"] and headers["Black"] not in self.playerList:
   self.updateSettingsList('Players', self.playerList, firstValue = headers["Black"])
  if '?' not in headers["Event"] and headers["Event"] not in self.eventList:
   self.updateSettingsList('Events', self.eventList, firstValue = headers["Event"])
  if '?' not in headers["Site"] and headers["Site"] not in self.siteList:
   self.updateSettingsList('Sites', self.siteList, firstValue = headers["Site"])
  self.saveSettings()
  self.game.headers = headers
  self._gameChanged = True

 @QtCore.pyqtSlot(chess.pgn.GameNode)
 def newGameNode(self, gameNode):
  self.notify('')
  self.gameNode = gameNode
  self._showEcoCode(self.game, fromBeginning = False)
  if gameNode.is_mainline():
   self.scorePlotGraphicsView.addGameNodes(gameNode)
   self.scorePlotGraphicsView.selectNodeItem(gameNode)
   if 'PlyCount' in self.game.headers:
    self.game.headers['PlyCount'] = str(gameNode.board().ply())
    self.gameHeaderTableView.setPlyCount(self.game.headers['PlyCount'])
   board = self.gameNode.board()
   if self.gameNode.is_end() and board.is_game_over():
    if board.is_checkmate():
     if board.turn == chess.WHITE:
      self.game.headers['Result'] = '0-1'
     else:
      self.game.headers['Result'] = '1-0'
    else:
     self.game.headers['Result'] = '1/2-1/2'
    self.gameChanged(self.game)

  if gameNode.is_main_variation():
   self.gameTreeViewWidget.addGameNodes(gameNode)
  else:
   self.gameTreeViewWidget.addVariant(gameNode)
  self._gameChanged = True
   
 @QtCore.pyqtSlot(chess.pgn.GameNode)
 def gameNodeSelected(self, gameNode):
  self.notify('')
  self.gameNode = gameNode
  self.boardGraphicsView.setGameNode(self.gameNode)
  self.gameTreeViewWidget.selectNodeItem(self.gameNode)
  self.scorePlotGraphicsView.selectNodeItem(self.gameNode)

 @QtCore.pyqtSlot(chess.pgn.Game)
 def gameChanged(self, game):
  self.notify('')
  self.gameHeaderTableView.setGameResult(game.headers['Result'])
  self.gameTreeViewWidget.setGameResult(game.headers['Result'])
  self.game = game
  self._gameChanged = True
  
class MzClassApplication(QtWidgets.QApplication):
 def __init__(self, argv : List[str], notifyFct : Callable[[str], None] = print) -> None: 
  super(MzClassApplication, self).__init__(argv)
  self.notifyFct = notifyFct

 def notify(self, rec, ev):
  rc = super(MzClassApplication, self).notify(rec, ev)
  self.notifyFct('{} -> Type(Event)= {}, handled = {}'.format(rec, ev.type(),  rc))
  return rc

 # ---------------------------------------------------------------------------

import os,  os.path

def runMzChess(notifyFct : Optional[Callable[[str], None]] = None):
 os.chdir(os.path.expanduser('~'))
 if notifyFct is not None:
  qApp = MzClassApplication(sys.argv)
 else:
  qApp = QtWidgets.QApplication(sys.argv)
 chessMainWindow = ChessMainWindow()
 chessMainWindow.show()
 chessMainWindow.setup()
 qApp.exec()

def _runMzChess():
 print('Hello, world')

if __name__ == "__main__":
 runMzChess(None)



import re
import sys, os.path

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import MzChess

if MzChess.useQt5:
 from PyQt5 import QtWidgets, QtGui
else:
 from PyQt6 import QtWidgets, QtGui

class UCIHighlighter(QtGui.QSyntaxHighlighter):
 sendKeywords = ['uci', 'debug', 'isready', 'register', 'ucinewgame', 'position', 'go', 'ponderhit', 'quit']
 receiveKeywords = ['id', 'uciok', 'isready', 'readyok', 'bestmove', 'registration', 'copyprotection', 'info', 'option']
 infoKeywords = ['depth', 'score', 'pv']
 
 def __init__(self, parent=None) -> None:
  super(UCIHighlighter, self).__init__(parent)
  sendFmt = self._charFormat(color = 'black', isBold = True,  isItalic = False)
  sendPattern = re.compile('(?<![a-z])({})(?![a-z])'.format('|'.join(self.sendKeywords)))
  receiveFmt = self._charFormat(color = 'blue', isBold = True,  isItalic = False)
  receivePattern = re.compile('(?<![a-z])({})(?![a-z])'.format('|'.join(self.receiveKeywords)))
  infoFmt = self._charFormat(color = 'green', isBold = True,  isItalic = False)
  infoPattern = re.compile('(?<![a-z])({})(?![a-z])'.format('|'.join(self.infoKeywords)))
  
  self.rules = [
   (sendPattern, sendFmt), 
   (receivePattern, receiveFmt), 
   (infoPattern, infoFmt)]

 def _charFormat(self, color : str = 'black', isBold : bool = False,  isItalic : bool = False) -> QtGui.QTextCharFormat:
  charFormat = QtGui.QTextCharFormat()
  _color = QtGui.QColor()
  _color.setNamedColor(color)
  charFormat.setForeground(_color)
  if isBold:
   charFormat.setFontWeight(QtGui.QFont.Weight.Bold)
  if isItalic:
   charFormat.setFontItalic(True)
  return charFormat
  
 def highlightBlock(self, text):
  for expr, charFormat in self.rules:
   for match in expr.finditer(text):
    self.setFormat(match.start(), match.end() - match.start(), charFormat)
  

class QUCIEdit(QtWidgets.QTextEdit):
 def __init__(self, parent = None) -> None:
  super(QUCIEdit, self).__init__(parent)
  
  self.uciHighlighter = UCIHighlighter(self.document())

if __name__ == "__main__":
 data = '''
ChessEngine/_toStdin: stdin> ucinewgame, readyok = TrueChessEngine/_toStdin: stdin> position fen r2qkbnr/pp2pppp/2n5/3P4/2pP4/2N2B2/PPP2PPP/R1BQK2R b KQkq - 0 7, readyok = TrueChessEngine/_toStdin: stdin> setoption name MultiPV value 1, readyok = TrueChessEngine/_toStdin: stdin> setoption name UCI_AnalyseMode value false, readyok = TrueChessEngine/_toStdin: stdin> go depth 15, readyok = TrueChessEngine/_fromStdout: stdout< info string NNUE evaluation using nn-82215d0fd0df.nnue enabled
ChessEngine/_fromStdout: stdout< info depth 1 seldepth 1 multipv 1 score cp -435 nodes 86 nps 86000 tbhits 0 time 1 pv c6b4
info depth 2 seldepth 2 multipv 1 score cp -415 nodes 120 nps 120000 tbhits 0 time 1 pv c6b4 a2a3
'''

 app = QtWidgets.QApplication([])
 editor = QUCIEdit()
 editor.setPlainText(data)
 editor.show()
 app.exec()
